# Copyright (c) 2016 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.api import extensions
from neutron.db import common_db_mixin
from neutron.db import db_base_plugin_v2
from neutron.db import dns_db
from neutron.db import extraroute_db
from neutron.db import l3_db
from neutron.db import l3_gwmode_db
from neutron.extensions import l3
from neutron.extensions import portbindings
from neutron.plugins.common import constants
from neutron.quota import resource_registry
from neutron_lib import exceptions
from oslo_log import log as logging
from oslo_utils import excutils
from sqlalchemy import inspect

from gbpservice._i18n import _LE
from gbpservice._i18n import _LI
from gbpservice.neutron import extensions as extensions_pkg
from gbpservice.neutron.extensions import cisco_apic_l3 as l3_ext
from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import (
    extension_db as extn_db)
from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import (
    mechanism_driver as md)

LOG = logging.getLogger(__name__)


class ApicL3Plugin(common_db_mixin.CommonDbMixin,
                   extraroute_db.ExtraRoute_db_mixin,
                   l3_gwmode_db.L3_NAT_db_mixin,
                   extn_db.ExtensionDbMixin,
                   dns_db.DNSDbMixin):

    supported_extension_aliases = ["router", "ext-gw-mode", "extraroute",
                                   "cisco-apic-l3", "dns-integration"]

    @staticmethod
    def get_plugin_type():
        return constants.L3_ROUTER_NAT

    @staticmethod
    def get_plugin_description():
        return _("L3 Router Service Plugin using the APIC via AIM")

    @resource_registry.tracked_resources(router=l3_db.Router,
                                         floatingip=l3_db.FloatingIP)
    def __init__(self):
        LOG.info(_LI("APIC AIM L3 Plugin __init__"))
        extensions.append_api_extensions_path(extensions_pkg.__path__)
        self._mechanism_driver = None
        super(ApicL3Plugin, self).__init__()

    @property
    def _md(self):
        if not self._mechanism_driver:
            # REVISIT(rkukura): It might be safer to search the MDs by
            # class rather than index by name, or to use a class
            # variable to find the instance.
            mech_mgr = self._core_plugin.mechanism_manager
            self._mechanism_driver = mech_mgr.mech_drivers['apic_aim'].obj
        return self._mechanism_driver

    def _extend_router_dict_apic(self, router_res, router_db):
        LOG.debug("APIC AIM L3 Plugin extending router dict: %s", router_res)
        session = inspect(router_db).session
        try:
            self._md.extend_router_dict(session, router_db, router_res)
            self._include_router_extn_attr(session, router_res)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("APIC AIM extend_router_dict failed"))

    db_base_plugin_v2.NeutronDbPluginV2.register_dict_extend_funcs(
        l3.ROUTERS, ['_extend_router_dict_apic'])

    def create_router(self, context, router):
        LOG.debug("APIC AIM L3 Plugin creating router: %s", router)
        self._md.ensure_tenant(context, router['router']['tenant_id'])
        with context.session.begin(subtransactions=True):
            # REVISIT(rkukura): The base operation may create a port,
            # which should generally not be done inside a
            # transaction. But we need to ensure atomicity, and are
            # generally not concerned with mechanism driver postcommit
            # processing. Consider overriding create_router_db()
            # instead, and/or reimplementing the base funtionality to
            # be completely transaction safe.
            result = super(ApicL3Plugin, self).create_router(context, router)
            self._process_router_op(context, result, router)
            self._md.create_router(context, result)
            return result

    def update_router(self, context, id, router):
        LOG.debug("APIC AIM L3 Plugin updating router %(id)s with: %(router)s",
                  {'id': id, 'router': router})
        with context.session.begin(subtransactions=True):
            # REVISIT(rkukura): The base operation sends notification
            # RPCs, which should generally not be done inside a
            # transaction. But we need to ensure atomicity, and are
            # not using an L3 agent. Consider overriding
            # create_router_db() instead, and/or reimplementing the
            # base funtionality to be completely transaction safe.
            original = self.get_router(context, id)
            result = super(ApicL3Plugin, self).update_router(context, id,
                                                             router)
            self._process_router_op(context, result, router)
            self._md.update_router(context, result, original)
            return result

    def delete_router(self, context, id):
        LOG.debug("APIC AIM L3 Plugin deleting router: %s", id)
        with context.session.begin(subtransactions=True):
            # REVISIT(rkukura): The base operation may delete ports
            # and sends notification RPCs, which should generally not
            # be done inside a transaction. But we need to ensure
            # atomicity, are not using an L3 agent, and are generally
            # not concerned with mechanism driver postcommit
            # processing. Consider reimplementing the base
            # funtionality to be completely transaction safe.
            router = self.get_router(context, id)
            super(ApicL3Plugin, self).delete_router(context, id)
            self._md.delete_router(context, router)

    def _process_router_op(self, context, result, router_req):
        self.set_router_extn_db(context.session, result['id'],
                                router_req['router'])
        self._include_router_extn_attr(context.session, result)

    def _include_router_extn_attr(self, session, router):
        attr = self.get_router_extn_db(session, router['id'])
        router.update(attr)

    def add_router_interface(self, context, router_id, interface_info):
        LOG.debug("APIC AIM L3 Plugin adding interface %(interface)s "
                  "to router %(router)s",
                  {'interface': interface_info, 'router': router_id})
        with context.session.begin(subtransactions=True):
            # REVISIT(rkukura): The base operation may create or
            # update a port and sends notification RPCs, which should
            # generally not be done inside a transaction. But we need
            # to ensure atomicity, are not using an L3 agent, and are
            # generally not concerned with mechanism driver postcommit
            # processing. Consider reimplementing the base
            # funtionality to be completely transaction safe.
            #
            # REVISIT: Remove override flag when no longer needed for
            # GBP.
            context.override_network_routing_topology_validation = (
                interface_info.get(
                    l3_ext.OVERRIDE_NETWORK_ROUTING_TOPOLOGY_VALIDATION))
            info = super(ApicL3Plugin, self).add_router_interface(
                context, router_id, interface_info)
            del context.override_network_routing_topology_validation
            # REVISIT(tbachman): This update port triggers port
            # binding, which means that port-binding happens inside
            # of a transaction, which shouldn't happen. This isn't
            # an issue with the AIM MD, but should be fixed at some
            # point (e.g. move port-binding outside, possibly queued
            # to be handled outside of the transaction, with some
            # sort of cleanup if it fails).
            self._core_plugin.update_port(context, info['port_id'],
                                          {'port': {portbindings.HOST_ID:
                                                    md.FABRIC_HOST_ID}})
            return info

    def _add_interface_by_subnet(self, context, router, subnet_id, owner):
        LOG.debug("APIC AIM L3 Plugin adding interface by subnet %(subnet)s "
                  "to router %(router)s",
                  {'subnet': subnet_id, 'router': router['id']})
        port, subnets, new_port = (
            super(ApicL3Plugin, self)._add_interface_by_subnet(
                context, router, subnet_id, owner))
        self._md.add_router_interface(context, router, port, subnets)
        return port, subnets, new_port

    def _add_interface_by_port(self, context, router, port_id, owner):
        LOG.debug("APIC AIM L3 Plugin adding interface by port %(port)s "
                  "to router %(router)s",
                  {'port': port_id, 'router': router['id']})
        port, subnets = (
            super(ApicL3Plugin, self)._add_interface_by_port(
                context, router, port_id, owner))
        self._md.add_router_interface(context, router, port, subnets)
        return port, subnets

    def remove_router_interface(self, context, router_id, interface_info):
        LOG.debug("APIC AIM L3 Plugin removing interface %(interface)s "
                  "from router %(router)s",
                  {'interface': interface_info, 'router': router_id})
        with context.session.begin(subtransactions=True):
            # REVISIT(rkukura): The base operation may delete or
            # update a port and sends notification RPCs, which should
            # generally not be done inside a transaction. But we need
            # to ensure atomicity, are not using an L3 agent, and are
            # generally not concerned with mechanism driver postcommit
            # processing. Consider reimplementing the base
            # funtionality to be completely transaction safe.
            info = super(ApicL3Plugin, self).remove_router_interface(
                context, router_id, interface_info)
            return info

    def _remove_interface_by_subnet(self, context, router_id, subnet_id,
                                    owner):
        LOG.debug("APIC AIM L3 Plugin removing interface by subnet %(subnet)s "
                  "from router %(router)s",
                  {'subnet': subnet_id, 'router': router_id})
        port_db, subnets = (
            super(ApicL3Plugin, self)._remove_interface_by_subnet(
                context, router_id, subnet_id, owner))
        self._md.remove_router_interface(context, router_id, port_db, subnets)
        return port_db, subnets

    def _remove_interface_by_port(self, context, router_id, port_id, subnet_id,
                                  owner):
        LOG.debug("APIC AIM L3 Plugin removing interface by port %(port)s "
                  "from router %(router)s",
                  {'port': port_id, 'router': router_id})
        port_db, subnets = (
            super(ApicL3Plugin, self)._remove_interface_by_port(
                context, router_id, port_id, subnet_id, owner))
        self._md.remove_router_interface(context, router_id, port_db, subnets)
        return port_db, subnets

    def create_floatingip(self, context, floatingip):
        fip = floatingip['floatingip']
        self._md.ensure_tenant(context, fip['tenant_id'])
        # Verify that subnet is not a SNAT host-pool
        self._md.check_floatingip_external_address(context, fip)
        with context.session.begin(subtransactions=True):
            if fip.get('subnet_id') or fip.get('floating_ip_address'):
                result = super(ApicL3Plugin, self).create_floatingip(
                    context, floatingip)
            else:
                # Iterate over non SNAT host-pool subnets and try to allocate
                # an address
                other_subs = self._md.get_subnets_for_fip(context, fip)
                result = None
                for ext_sn in other_subs:
                    fip['subnet_id'] = ext_sn
                    try:
                        with context.session.begin(nested=True):
                            result = (super(ApicL3Plugin, self)
                                      .create_floatingip(context, floatingip))
                        break
                    except exceptions.IpAddressGenerationFailure:
                        LOG.info(_LI('No more floating IP addresses available '
                                     'in subnet %s'),
                                 ext_sn)

                if not result:
                    raise exceptions.IpAddressGenerationFailure(
                        net_id=fip['floating_network_id'])
            self._md.create_floatingip(context, result)
            self.update_floatingip_status(context, result['id'],
                                          result['status'])
        return result

    def update_floatingip(self, context, id, floatingip):
        with context.session.begin(subtransactions=True):
            old_fip = self.get_floatingip(context, id)
            result = super(ApicL3Plugin, self).update_floatingip(
                context, id, floatingip)
            self._md.update_floatingip(context, old_fip, result)
            if old_fip['status'] != result['status']:
                self.update_floatingip_status(context, result['id'],
                                              result['status'])
        return result

    def delete_floatingip(self, context, id):
        with context.session.begin(subtransactions=True):
            old_fip = self.get_floatingip(context, id)
            super(ApicL3Plugin, self).delete_floatingip(context, id)
            self._md.delete_floatingip(context, old_fip)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import commands
import logging
import sys

from libs import config_libs
from libs import utils_libs
from libs import verify_libs


def main():

    # Run the Testcases:
    test = test_gbp_prs_neg()
    if test.test_gbp_prs_neg_1() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PRS_NEG_1')
    if test.test_gbp_prs_neg_2() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PRS_NEG_2')
    if test.test_gbp_prs_neg_3() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PRS_NEG_3')
    test.cleanup()
    utils_libs.report_results('test_gbp_prs_neg', 'test_results.txt')
    sys.exit(1)


class test_gbp_prs_neg(object):

    # Initialize logging
    logging.basicConfig(
        format='%(asctime)s [%(levelname)s] %(name)s - %(message)s',
        level=logging.WARNING)
    _log = logging.getLogger(__name__)
    cmd = 'rm /tmp/test_gbp_prs_neg.log'
    commands.getoutput(cmd)
    hdlr = logging.FileHandler('/tmp/test_gbp_prs_neg.log')
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    hdlr.setFormatter(formatter)
    _log.addHandler(hdlr)
    _log.setLevel(logging.INFO)
    _log.setLevel(logging.DEBUG)

    def __init__(self):
        """
        Init def
        """
        self._log.info(
            "\n## START OF GBP POLICY_RULE_SET NEGATIVE TESTSUITE\n")
        self.gbpcfg = config_libs.Gbp_Config()
        self.gbpverify = verify_libs.Gbp_Verify()
        self.act_name = 'demo_pa'
        self.cls_name = 'demo_pc'
        self.rule_name = 'demo_pr'
        self.ruleset_name = 'demo_prs'
        self._log.info('\n## Step 1: Create a PC needed for PRS Testing ##')
        self.cls_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'classifier', self.cls_name)
        if self.cls_uuid == 0:
            self._log.info(
                "\nReqd Policy Classifier Create Failed, "
                "hence GBP Policy Rule-Set Negative Test Suite Run ABORTED\n")
            return
        self._log.info('\n## Step 1: Create a PA needed for PRS Testing ##')
        self.act_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'action', self.act_name)
        if self.act_uuid == 0:
            self._log.info(
                "\nReqd Policy Action Create Failed, hence GBP "
                "Policy Rule-Set Negative Test Suite Run ABORTED\n")
            return
        self._log.info('\n## Step 1: Create a PR needed for PRS Testing ##')
        self.rule_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'rule', self.rule_name, classifier=self.cls_name,
            action=self.act_name)
        if self.rule_uuid == 0:
            self._log.info(
                "\nReqd Policy Rule Create Failed, hence GBP "
                "Policy Rule-Set Negative Test Suite Run ABORTED\n ")
            return

    def cleanup(self, tc_name=''):
        if tc_name != '':
            self._log.info('%s: FAILED' % (tc_name))
        for obj in ['ruleset', 'rule', 'classifier', 'action']:
            self.gbpcfg.gbp_del_all_anyobj(obj)

    def test_gbp_prs_neg_1(self):

        self._log.info(
            "\n#################################################\n"
            "TESTCASE_GBP_PRS_NEG_1: TO CREATE/VERIFY POLICY "
            "RULESET with INVALID POLICY RULE\n"
            "TEST_STEPS::\n"
            "Create Policy RuleSet Object with Invalid PR\n"
            "Verify the create FAILs and config rolls back\n"
            "#################################################\n")

        # Testcase work-flow starts
        self._log.info("\n## Step 1: Create Policy RuleSet with Invalid PR ##")
        if self.gbpcfg.gbp_policy_cfg_all(
                1,
                'ruleset',
                self.ruleset_name,
                policy_rules='INVALID') != 0:
            self._log.info(
                "# Step 1: Create RuleSet with Invalid PR did NOT Fail")
            return 0
        self._log.info('# Step 1A: Verify Policy RuleSet config rolled back')
        if self.gbpverify.gbp_policy_verify_all(
                1, 'ruleset', self.ruleset_name) != 0:
            self._log.info(
                "# Step 1A: Verify RuleSet config roll back did NOT Fail")
            return 0
        self._log.info("\nTESTCASE_GBP_PRS_NEG_1: PASSED")
        return 1

    def test_gbp_prs_neg_2(self):

        self._log.info(
            "\n###################################################\n"
            "TESTCASE_GBP_PRS_NEG_2: TO CREATE/VERIFY POLICY "
            "RULESET with mix of VALID and  INVALID POLICY RULE\n"
            "TEST_STEPS::\n"
            "Create Policy RuleSet with a mix of Valid and Invalid PR\n"
            "Verify the create FAILs and config rolls back\n"
            "##################################################\n")

        # Testcase work-flow starts
        self._log.info(
            "\n## Step 1: Create Policy RuleSet with mix of Valid "
            "and Invalid PR ##")
        if self.gbpcfg.gbp_policy_cfg_all(
                1,
                'ruleset',
                self.ruleset_name,
                policy_rules="'%s INVALID'" %
                (self.rule_uuid)) != 0:
            self._log.info(
                "# Step 1: Create RuleSet with mix of Valid and "
                "Invalid PR did NOT Fail")
            return 0
        self._log.info('# Step 1A: Verify Policy RuleSet config rolled back')
        if self.gbpverify.gbp_policy_verify_all(
                1, 'ruleset', self.ruleset_name) != 0:
            self._log.info(
                "# Step 1A: Verify RuleSet config roll back did NOT Fail")
            return 0
        self._log.info("\nTESTCASE_GBP_PRS_NEG_2: PASSED")
        return 1

    def test_gbp_prs_neg_3(self):

        self._log.info(
            "\n###################################################\n"
            "TESTCASE_GBP_PRS_NEG_3: TO UPDATE/VERIFY POLICY "
            "RULE with VALID and INVALID PR\n"
            "TEST_STEPS::\n"
            "Create a Policy RuleSet with default attribute\n"
            "Update the Policy RuleSet with a mix of Valid and Invalid PR\n"
            "Verify the update fails and config roll backs to "
            "original values of the PRS\n"
            "##################################################\n")

        # Testcase work-flow starts
        self._log.info('\n## Step 1: Create a PRS with default attribute ##\n')
        ruleset_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'ruleset', self.ruleset_name)
        if ruleset_uuid == 0:
            self._log.info(
                "\n## Step 1: Create RuleSet with default attr == Failed")
            return 0
        self._log.info(
            "\n## Step 2: Update the PRS with VALID PR and INVALID PR")
        if self.gbpcfg.gbp_policy_cfg_all(
                2, 'ruleset', ruleset_uuid, policy_rule='"%s INVALID"' %
                (self.rule_name)) != 0:
            self._log.info(
                "\n## Step 2: Updating Policy RuleSet with VALID "
                "and INVALID Policy Rules did NOT Fail")
            return 0
        self._log.info(
            '# Step 2A: Verify RuleSet config update has been rolled back')
        if self.gbpverify.gbp_policy_verify_all(
                1,
                'ruleset',
                self.ruleset_name,
                id=ruleset_uuid,
                policy_rules=self.rule_uuid,
                shared='False') != 0:
            self._log.info("# Step 2A: Verify RuleSet roll back did NOT Fail")
            return 0
        self._log.info("\nTESTCASE_GBP_PRS_NEG_3: PASSED")
        return 1

if __name__ == '__main__':
    main()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg as oslo_config
from oslo_log import log as oslo_logging
from oslo_utils import importutils

from gbpservice.nfp.core import context

import logging
import os
import sys

EVENT = 50
logging.addLevelName(EVENT, "EVENT")
CONF = oslo_config.CONF


if hasattr(sys, 'frozen'):  # support for py2exe
    _srcfile = "logging%s__init__%s" % (os.sep, __file__[-4:])
elif __file__[-4:].lower() in ['.pyc', '.pyo']:
    _srcfile = __file__[:-4] + '.py'
else:
    _srcfile = __file__
_srcfile = os.path.normcase(_srcfile)


def currentframe():
    """Return the frame object for the caller's stack frame."""
    try:
        raise Exception
    except Exception:
        return sys.exc_info()[2].tb_frame.f_back


if hasattr(sys, '_getframe'):
    currentframe = lambda: sys._getframe(3)


class NfpLogAdapter(oslo_logging.KeywordArgumentAdapter):

    def event(self, msg, *args, **kwargs):
        self.log(EVENT, msg, *args, **kwargs)


class WrappedLogger(logging.Logger):

    def __init__(self, name):
        logging.Logger.__init__(self, name)

    def findCaller(self):
        """
        Find the stack frame of the caller so that we can note the source
        file name, line number and function name.
        """
        f = currentframe()
        # On some versions of IronPython, currentframe() returns None if
        # IronPython isn't run with -X:Frames.
        if f is not None:
            f = f.f_back
            if f.f_back:
                f = f.f_back
        rv = "(unknown file)", 0, "(unknown function)"
        while hasattr(f, "f_code"):
            co = f.f_code
            filename = os.path.normcase(co.co_filename)
            if filename == _srcfile:
                f = f.f_back
                continue
            rv = (co.co_filename, f.f_lineno, co.co_name)
            break
        return rv

    def _get_nfp_msg(self, msg):
        nfp_context = context.get()
        log_context = nfp_context['log_context']
        if log_context:
            ctxt = "[%s] [NFI:%s] [NFD:%s]" % (log_context.get(
                'meta_id', '-'),
                log_context.get('nfi_id', '-'),
                log_context.get('nfd_id', '-'))
            msg = "%s %s" % (ctxt, msg)

        component = ''
        if hasattr(CONF, 'module'):
            component = CONF.module
        msg = "[%s] %s" % (component, msg)
        return msg

    def makeRecord(self, name, level, fn,
                   lno, msg, args, exc_info, func=None, extra=None):
        # Prefix log meta id with every log if project is 'nfp'
        if extra and extra.get('project') == 'nfp':
            msg = self._get_nfp_msg(msg)
        return super(WrappedLogger, self).makeRecord(
            name, level, fn, lno, msg,
            args, exc_info, func=func, extra=extra)


def init_logger(logger_class):
    logging.setLoggerClass(importutils.import_class(logger_class))


def getLogger(name, **kwargs):
    kwargs.update(project='nfp')
    logger = NfpLogAdapter(logging.getLogger(name),
                           kwargs)
    return logger



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import pdb
import sys

"""For debugging inside a child process.

    import pdb;pdb.set_trace() does not work
    with python multiprocessing.
    Instead use below pdb class to debug inside
    a worker process / child process.
"""


class ForkedPdb(pdb.Pdb):

    """A Pdb subclass that may be used
    from a forked multiprocessing child

    """

    def interaction(self, *args, **kwargs):
        _stdin = sys.stdin
        try:
            sys.stdin = file('/dev/stdin')
            pdb.Pdb.interaction(self, *args, **kwargs)
        finally:
            sys.stdin = _stdin



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from gbpservice.nfp.core import log as nfp_logging

import collections
import six

deque = collections.deque

LOG = nfp_logging.getLogger(__name__)


class Supress(object):

    def __init__(self, ignore_list=None):
        self._ignore = ignore_list or []

    def __enter__(self):
        pass

    def __exit__(self, e_type, e_value, traceback):
        if e_type in self._ignore:
            return True
        for exception in self._ignore:
            if isinstance(e_type, exception):
                return True


class Path(object):

    def __init__(self, name):
        self._waitq = deque()
        self.name = name
        self.count = 0
        self.invalid = False

    def queue(self, event):
        self._waitq.append(event)

    def pop(self):
        events = []

        with Supress([IndexError]):
            events.append(self._waitq.popleft())
        return events

    def done(self):
        self._waitq.clear()

# {'key': {'current':Path, 'waiting':Path}
paths = {}


def run():
    for key, path in six.iteritems(paths):
        if path['current'].count == 0:
            path['current'].done()
            if path['waiting'].name != 'INVALID':
                path['current'] = path['waiting']
                path['current'].invalid = False
                path['waiting'] = Path('INVALID')

    events = []
    # Get any queued events in the current path
    for key, path in six.iteritems(paths):
        events += path['current'].pop()
    return events


def event_complete(event):
    name = event.desc.path_type
    key = event.desc.path_key

    if not name:
        return
    name = name.upper()
    with Supress([KeyError]):
        path = paths[key]
        if path['current'].name != name:
            return
        path['current'].count -= 1


def schedule_event(event):
    name = event.desc.path_type
    key = event.desc.path_key

    if not name:
        return 'schedule'

    name = name.upper()

    try:
        path = paths[key]
        if path['current'].name == name:
            if path['current'].invalid:
                return 'discard'
            path['current'].count += 1
            return 'schedule'

        if path['waiting'].name == name:
            path['waiting'].queue(event)
            return 'wait'

        if path['current'].name != name:
            return 'discard'
    except Exception:
        return 'schedule'
    return 'schedule'


def path_complete(path_type, key):
    try:
        path = paths[key]
        if path['current'].name == path_type.upper() and (
                path['waiting'].name == 'INVALID'):
            paths.pop(key)
    except KeyError:
        message = "Path completion - %s path does not exist with key %s" % (
            path_type, key)
        LOG.debug(message)


def create_path(key):
        # Create cannot progress if there is already a path
        # with the same key in any state
    try:
        path = paths[key]
        assert False, "Path (%s) with key (%s) is already in progress" % (
            path['current'].name, key)
    except KeyError:
        # Create new path
        paths[key] = {'current': Path('CREATE'), 'waiting': Path('INVALID')}


def delete_path(key):
    try:
        path = paths[key]
        if path['current'].name != 'DELETE':
            path['waiting'] = Path('DELETE')
            path['current'].invalid = True
        else:
            assert False, ("Delete Path (%s) with key (%s)"
                           "is already in progress" % (
                               path['current'].name, key))
    except KeyError:
        paths[key] = {'current': Path('DELETE'), 'waiting': Path('INVALID')}


def update_path(key):
    # Update cannot progress if there is DELETE already in progress
    # or DELETE already waiting.
    try:
        path = paths[key]
        assert False, "Path (%s) with key (%s) is in progress" % (
            path.name, key)
    except KeyError:
        # Create new path
        paths[key] = {'current': Path('UPDATE'), 'waiting': Path('INVALID')}



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""gbp_security_group_mapping

Revision ID: 5c65abe72596
Revises: 3ef186997b02
Create Date: 2014-08-21 14:30:23.68888

"""

# revision identifiers, used by Alembic.
revision = '5c65abe72596'
down_revision = '3ef186997b02'

from alembic import op
import sqlalchemy as sa


def upgrade():

    op.create_table(
        'gpm_policy_rule_set_sg_mapping',
        sa.Column('policy_rule_set_id', sa.String(length=36), nullable=False),
        sa.Column('provided_sg_id', sa.String(length=36)),
        sa.Column('consumed_sg_id', sa.String(length=36)),
        sa.ForeignKeyConstraint(['policy_rule_set_id'],
                                ['gp_policy_rule_sets.id'],
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['provided_sg_id'], ['securitygroups.id']),
        sa.ForeignKeyConstraint(['consumed_sg_id'], ['securitygroups.id']),
        sa.PrimaryKeyConstraint('policy_rule_set_id')
    )


def downgrade():

    op.drop_table('gpm_policy_rule_set_sg_mapping')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from apic_ml2.neutron.db import port_ha_ipaddress_binding as ha_ip_db
import mock
import netaddr
from neutron.common import config  # noqa
from neutron import context
from neutron import manager
from oslo_config import cfg
import unittest2

from gbpservice.neutron.services.grouppolicy import (
    policy_driver_manager as pdm)
from gbpservice.neutron.services.servicechain.plugins.ncp import (
    plugin as ncp_plugin)
from gbpservice.neutron.services.servicechain.plugins.ncp import model
from gbpservice.neutron.tests.unit.services.grouppolicy import (
    test_apic_mapping as test_apic)
from gbpservice.neutron.tests.unit.services.grouppolicy import (
    test_resource_mapping as test_rmd)
from gbpservice.neutron.tests.unit.services.servicechain.ncp import (
    test_ncp_plugin as base)
from gbpservice.neutron.tests.unit.services.servicechain.ncp import (
    test_tscp_resource_mapping as test_tscp_rmd)


class ApicMappingStitchingPlumberGBPTestCase(
        test_apic.ApicMappingTestCase):

    def setUp(self, plumber='stitching_plumber'):
        cfg.CONF.set_override(
            'extension_drivers', ['apic_segmentation_label',
                                  'proxy_group'], group='group_policy')
        cfg.CONF.set_override('node_plumber', plumber,
                              group='node_composition_plugin')
        super(ApicMappingStitchingPlumberGBPTestCase, self).setUp(
            sc_plugin=base.SC_PLUGIN_KLASS)

        def get_plumbing_info(context):
            return test_tscp_rmd.info_mapping.get(
                context.current_profile['service_type'])

        self.node_driver = self.sc_plugin.driver_manager.ordered_drivers[0].obj
        self.node_driver.get_plumbing_info = get_plumbing_info
        self.mgr = self.driver.apic_manager

        def get_plumbing_info(context):
            return test_tscp_rmd.info_mapping.get(
                context.current_profile['service_type'])

        self.node_driver = self.sc_plugin.driver_manager.ordered_drivers[0].obj
        self.node_driver.get_plumbing_info = get_plumbing_info
        self.saved_get_policy_target_group_status = (
            pdm.PolicyDriverManager.get_policy_target_group_status)
        pdm.PolicyDriverManager.get_policy_target_group_status = (
                mock.MagicMock({}))

    def tearDown(self):
        pdm.PolicyDriverManager.get_policy_target_group_status = (
                self.saved_get_policy_target_group_status)
        super(ApicMappingStitchingPlumberGBPTestCase, self).tearDown()

    @property
    def sc_plugin(self):
        plugins = manager.NeutronManager.get_service_plugins()
        servicechain_plugin = plugins.get('SERVICECHAIN')
        return servicechain_plugin


class TestPolicyRuleSet(ApicMappingStitchingPlumberGBPTestCase,
                        test_apic.TestPolicyRuleSet):
    pass


class TestPolicyRule(ApicMappingStitchingPlumberGBPTestCase,
                     test_apic.TestPolicyRule):
    pass


class TestExternalSegment(ApicMappingStitchingPlumberGBPTestCase,
                          test_apic.TestExternalSegment):
    pass


class TestExternalPolicy(ApicMappingStitchingPlumberGBPTestCase,
                         test_apic.TestExternalPolicy):
    pass


class TestApicChains(ApicMappingStitchingPlumberGBPTestCase,
                     base.NodeCompositionPluginTestMixin):

    def _assert_proper_chain_instance(self, sc_instance, provider_ptg_id,
                                      policy_rule_set_id, scs_id_list,
                                      classifier_id=None):
        self.assertEqual(sc_instance['provider_ptg_id'], provider_ptg_id)
        self.assertEqual(sc_instance['consumer_ptg_id'], 'N/A')
        self.assertEqual(scs_id_list, sc_instance['servicechain_specs'])
        provider = self.show_policy_target_group(
            provider_ptg_id)['policy_target_group']
        self.assertEqual(sc_instance['tenant_id'], provider['tenant_id'])
        if classifier_id:
            self.assertEqual(sc_instance['classifier_id'], classifier_id)

    def test_classifier_update_to_chain(self):
        scs_id = self._create_servicechain_spec(node_types=['FIREWALL'])
        _, classifier_id, policy_rule_id = self._create_tcp_redirect_rule(
                                                            "20:90", scs_id)

        policy_rule_set = self.create_policy_rule_set(
            name="c1", policy_rules=[policy_rule_id])
        policy_rule_set_id = policy_rule_set['policy_rule_set']['id']
        provider_ptg_id, consumer_ptg_id = self._create_provider_consumer_ptgs(
                                                            policy_rule_set_id)

        sc_instances = self._list_service_chains()
        # We should have one service chain instance created now
        self.assertEqual(len(sc_instances['servicechain_instances']), 1)
        sc_instance = sc_instances['servicechain_instances'][0]
        self._assert_proper_chain_instance(sc_instance, provider_ptg_id,
                                           consumer_ptg_id, [scs_id])
        with mock.patch.object(
                ncp_plugin.NodeCompositionPlugin,
                'notify_chain_parameters_updated') as notify_chain_update:

            # Update classifier and verify instance is updated
            self.update_policy_classifier(classifier_id, port_range=80)

            notify_chain_update.assert_called_once_with(
                mock.ANY, sc_instance['id'])
            sc_instances = self._list_service_chains()
            self.assertEqual(len(sc_instances['servicechain_instances']), 1)
            sc_instance_updated = sc_instances['servicechain_instances'][0]
            self.assertEqual(sc_instance, sc_instance_updated)

        self._verify_ptg_delete_cleanup_chain(provider_ptg_id)

    def test_redirect_multiple_ptgs_single_prs(self):
        scs_id = self._create_servicechain_spec()
        _, _, policy_rule_id = self._create_tcp_redirect_rule(
                                                "20:90", scs_id)

        policy_rule_set = self.create_policy_rule_set(
            name="c1", policy_rules=[policy_rule_id])
        policy_rule_set_id = policy_rule_set['policy_rule_set']['id']

        #Create 2 provider and 2 consumer PTGs
        provider_ptg1 = self.create_policy_target_group(
            name="p_ptg1",
            provided_policy_rule_sets={policy_rule_set_id: None})
        provider_ptg1_id = provider_ptg1['policy_target_group']['id']
        self.create_policy_target_group(
            name="c_ptg1",
            consumed_policy_rule_sets={policy_rule_set_id: None})

        provider_ptg2 = self.create_policy_target_group(
            name="p_ptg2",
            provided_policy_rule_sets={policy_rule_set_id: None})
        provider_ptg2_id = provider_ptg2['policy_target_group']['id']
        self.create_policy_target_group(
            name="c_ptg2",
            consumed_policy_rule_sets={policy_rule_set_id: None})

        sc_instances = self._list_service_chains()
        # We should have 2 service chain instances (one per provider)
        self.assertEqual(len(sc_instances['servicechain_instances']), 2)
        sc_instances = sc_instances['servicechain_instances']
        sc_instances_provider_ptg_ids = set()
        sc_instances_consumer_ptg_ids = set()
        for sc_instance in sc_instances:
            sc_instances_provider_ptg_ids.add(sc_instance['provider_ptg_id'])
            sc_instances_consumer_ptg_ids.add(sc_instance['consumer_ptg_id'])
        expected_provider_ptg_ids = {provider_ptg1_id, provider_ptg2_id}
        self.assertEqual(expected_provider_ptg_ids,
                         sc_instances_provider_ptg_ids)

        # Deleting one provider should end up deleting the one service chain
        # Instance associated to it
        self.delete_policy_target_group(
            provider_ptg1_id, expected_res_status=204)

        sc_instances = self._list_service_chains()
        self.assertEqual(len(sc_instances['servicechain_instances']), 1)
        sc_instance = sc_instances['servicechain_instances'][0]
        self.assertNotEqual(sc_instance['provider_ptg_id'], provider_ptg1_id)

        self.delete_policy_target_group(
            provider_ptg2_id, expected_res_status=204)

        sc_instances = self._list_service_chains()
        # No more service chain instances when all the providers are deleted
        self.assertEqual(len(sc_instances['servicechain_instances']), 0)

    def test_redirect_to_chain(self):
        scs_id = self._create_servicechain_spec()
        _, _, policy_rule_id = self._create_tcp_redirect_rule(
                                                "20:90", scs_id)

        policy_rule_set = self.create_policy_rule_set(
            name="c1", policy_rules=[policy_rule_id])
        policy_rule_set_id = policy_rule_set['policy_rule_set']['id']
        provider_ptg_id, consumer_ptg_id = self._create_provider_consumer_ptgs(
                                                            policy_rule_set_id)

        sc_instances = self._list_service_chains()
        # We should have one service chain instance created now
        self.assertEqual(len(sc_instances['servicechain_instances']), 1)

        # Verify that PTG delete cleans up the chain instances
        self.delete_policy_target_group(
            provider_ptg_id, expected_res_status=204)

        sc_instances = self._list_service_chains()
        self.assertEqual(len(sc_instances['servicechain_instances']), 0)

    def test_rule_update_updates_chain(self):
        scs_id = self._create_servicechain_spec()
        _, _, policy_rule_id = self._create_tcp_redirect_rule("20:90", scs_id)

        policy_rule_set = self.create_policy_rule_set(
            name="c1", policy_rules=[policy_rule_id])
        policy_rule_set_id = policy_rule_set['policy_rule_set']['id']
        provider_ptg_id, consumer_ptg_id = self._create_provider_consumer_ptgs(
                                                            policy_rule_set_id)
        sc_instances = self._list_service_chains()
        # We should have one service chain instance created now
        self.assertEqual(len(sc_instances['servicechain_instances']), 1)
        sc_instance = sc_instances['servicechain_instances'][0]

        # Update policy rule with new classifier and verify instance is
        # recreated
        classifier = self.create_policy_classifier(
            protocol='TCP', port_range="80",
            direction='bi')['policy_classifier']

        self.update_policy_rule(policy_rule_id,
                                policy_classifier_id=classifier['id'])
        sc_instances = self._list_service_chains()
        # We should have one service chain instance created now
        self.assertEqual(len(sc_instances['servicechain_instances']), 1)
        sc_instance_new = sc_instances['servicechain_instances'][0]
        self.assertNotEqual(sc_instance, sc_instance_new)

        scs_id2 = self._create_servicechain_spec()
        action = self.create_policy_action(
            action_type='redirect', action_value=scs_id2)['policy_action']
        self.update_policy_rule(policy_rule_id, policy_actions=[action['id']])

        # Verify SC instance changed
        sc_instances = self._list_service_chains()
        # We should have one service chain instance created now
        self.assertEqual(len(sc_instances['servicechain_instances']), 1)
        sc_instance_new = sc_instances['servicechain_instances'][0]
        self.assertNotEqual(sc_instance, sc_instance_new)

        self.delete_policy_target_group(
            provider_ptg_id, expected_res_status=204)

        sc_instances = self._list_service_chains()
        self.assertEqual(len(sc_instances['servicechain_instances']), 0)

    def test_update_ptg_with_redirect_prs(self):
        scs_id = self._create_servicechain_spec()
        _, _, policy_rule_id = self._create_tcp_redirect_rule(
                                                "20:90", scs_id)

        policy_rule_set = self.create_policy_rule_set(
            name="c1", policy_rules=[policy_rule_id])
        policy_rule_set_id = policy_rule_set['policy_rule_set']['id']
        provider_ptg, consumer_ptg = self._create_provider_consumer_ptgs()

        sc_instances = self._list_service_chains()
        self.assertEqual(len(sc_instances['servicechain_instances']), 0)

        # We should have one service chain instance created when PTGs are
        # updated with provided and consumed prs
        self.update_policy_target_group(
                            provider_ptg,
                            provided_policy_rule_sets={policy_rule_set_id: ''},
                            consumed_policy_rule_sets={},
                            expected_res_status=200)
        self.update_policy_target_group(
                            consumer_ptg,
                            provided_policy_rule_sets={},
                            consumed_policy_rule_sets={policy_rule_set_id: ''},
                            expected_res_status=200)

        sc_instances = self._list_service_chains()
        self.assertEqual(len(sc_instances['servicechain_instances']), 1)

        # Verify that PTG update removing prs cleans up the chain instances
        self.update_policy_target_group(
            provider_ptg, provided_policy_rule_sets={},
            consumed_policy_rule_sets={}, expected_res_status=200)

        sc_instances = self._list_service_chains()
        self.assertEqual(len(sc_instances['servicechain_instances']), 0)

    def test_chain_on_apic_create(self):
        scs_id = self._create_servicechain_spec(
            node_types=['FIREWALL_TRANSPARENT'])
        _, _, policy_rule_id = self._create_tcp_redirect_rule(
            "20:90", scs_id)
        policy_rule_set = self.create_policy_rule_set(
            name="c1", policy_rules=[policy_rule_id])['policy_rule_set']
        # Create PTGs on same L2P
        l2p = self.create_l2_policy()['l2_policy']
        provider = self.create_policy_target_group(
            l2_policy_id=l2p['id'])['policy_target_group']
        consumer = self.create_policy_target_group(
            l2_policy_id=l2p['id'])['policy_target_group']

        mgr = self.driver.apic_manager
        mgr.reset_mock()
        # Provide the redirect contract
        self.update_policy_target_group(
            consumer['id'],
            consumed_policy_rule_sets={policy_rule_set['id']: ''})
        self.assertFalse(mgr.ensure_bd_created_on_apic.called)
        self.assertFalse(mgr.ensure_epg_created.called)

        # Now form the chain
        self.update_policy_target_group(
            provider['id'],
            provided_policy_rule_sets={policy_rule_set['id']: ''})

        sc_instances = self._list_service_chains()
        # We should have one service chain instance created now
        self.assertEqual(len(sc_instances['servicechain_instances']), 1)
        sc_instance = sc_instances['servicechain_instances'][0]

        expected = [
            # Provider EPG provided PRS
            mock.call(provider['tenant_id'], provider['id'],
                      policy_rule_set['id'], provider=True,
                      contract_owner=policy_rule_set['tenant_id'],
                      transaction=mock.ANY),
            # Consumer EPG consumed PRS
            mock.call(consumer['tenant_id'], consumer['id'],
                      policy_rule_set['id'], provider=False,
                      contract_owner=policy_rule_set['tenant_id'],
                      transaction=mock.ANY)]

        self._verify_chain_set(provider, l2p, policy_rule_set,
                               sc_instance, 1, pre_set_contract_calls=expected)

        # New consumer doesn't trigger anything
        new_consumer = self.create_policy_target_group(
            l2_policy_id=l2p['id'])['policy_target_group']
        mgr.reset_mock()
        self.update_policy_target_group(
            new_consumer['id'],
            consumed_policy_rule_sets={policy_rule_set['id']: ''})

        self.assertFalse(mgr.ensure_bd_created_on_apic.called)
        self.assertFalse(mgr.ensure_epg_created.called)

    def test_chain_on_apic_create_shared(self):
        scs_id = self._create_servicechain_spec(
            node_types=['FIREWALL_TRANSPARENT'], shared=True)
        policy_rule_id = self._create_simple_policy_rule(
            action_type='redirect', shared=True, action_value=scs_id)['id']
        policy_rule_set = self.create_policy_rule_set(
            name="c1", policy_rules=[policy_rule_id],
            shared=True)['policy_rule_set']
        # Create PTGs on same L2P
        l2p = self.create_l2_policy(shared=True,
                                    tenant_id='admin')['l2_policy']
        provider = self.create_policy_target_group(
            l2_policy_id=l2p['id'], shared=True)['policy_target_group']
        consumer = self.create_policy_target_group(
            l2_policy_id=l2p['id'])['policy_target_group']

        mgr = self.driver.apic_manager
        mgr.reset_mock()
        # Provide the redirect contract
        self.update_policy_target_group(
            consumer['id'],
            consumed_policy_rule_sets={policy_rule_set['id']: ''})
        self.assertFalse(mgr.ensure_bd_created_on_apic.called)
        self.assertFalse(mgr.ensure_epg_created.called)

        # Now form the chain
        self.update_policy_target_group(
            provider['id'],
            provided_policy_rule_sets={policy_rule_set['id']: ''})

        sc_instances = self._list_service_chains()
        # We should have one service chain instance created now
        self.assertEqual(len(sc_instances['servicechain_instances']), 1)
        sc_instance = sc_instances['servicechain_instances'][0]

        expected = [
            # Provider EPG provided PRS
            mock.call('common', provider['id'],
                      policy_rule_set['id'], provider=True,
                      contract_owner='common',
                      transaction=mock.ANY),
            # Consumer EPG consumed PRS
            mock.call(consumer['tenant_id'], consumer['id'],
                      policy_rule_set['id'], provider=False,
                      contract_owner='common',
                      transaction=mock.ANY)]

        self._verify_chain_set(provider, l2p, policy_rule_set,
                               sc_instance, 1, pre_set_contract_calls=expected)

        # New consumer doesn't trigger anything
        new_consumer = self.create_policy_target_group(
            l2_policy_id=l2p['id'])['policy_target_group']
        mgr.reset_mock()
        self.update_policy_target_group(
            new_consumer['id'],
            consumed_policy_rule_sets={policy_rule_set['id']: ''})

        self.assertFalse(mgr.ensure_bd_created_on_apic.called)
        self.assertFalse(mgr.ensure_epg_created.called)
        mgr.reset_mock()
        self.update_policy_target_group(
            provider['id'], provided_policy_rule_sets={})
        # Provider EPG contract unset
        expected = mock.call('common', provider['id'],
                             policy_rule_set['id'], provider=True,
                             contract_owner='common',
                             transaction=mock.ANY)
        self._verify_chain_unset(
            provider, l2p, policy_rule_set,
            sc_instance, 1, pre_unset_contract_calls=[expected])

    def test_chain_on_apic_delete(self):
        scs_id = self._create_servicechain_spec(
            node_types=['FIREWALL_TRANSPARENT', 'FIREWALL_TRANSPARENT'])
        _, _, policy_rule_id = self._create_tcp_redirect_rule(
            "20:90", scs_id)
        policy_rule_set = self.create_policy_rule_set(
            name="c1", policy_rules=[policy_rule_id])['policy_rule_set']
        # Create PTGs on same L2P
        l2p = self.create_l2_policy()['l2_policy']
        provider = self.create_policy_target_group(
            l2_policy_id=l2p['id'])['policy_target_group']
        consumer = self.create_policy_target_group(
            l2_policy_id=l2p['id'])['policy_target_group']

        # Provide the redirect contract
        self.update_policy_target_group(
            provider['id'],
            provided_policy_rule_sets={policy_rule_set['id']: ''})

        # Now form the chain
        self.update_policy_target_group(
            consumer['id'],
            consumed_policy_rule_sets={policy_rule_set['id']: ''})

        sc_instances = self._list_service_chains()
        # We should have one service chain instance created now
        sc_instance = sc_instances['servicechain_instances'][0]

        # Dissolve the chain by disassociation
        mgr = self.driver.apic_manager
        mgr.reset_mock()
        self.update_policy_target_group(
            provider['id'], provided_policy_rule_sets={})
        # Provider EPG contract unset
        expected = mock.call(provider['tenant_id'], provider['id'],
                             policy_rule_set['id'], provider=True,
                             contract_owner=policy_rule_set['tenant_id'],
                             transaction=mock.ANY)
        self._verify_chain_unset(
            provider, l2p, policy_rule_set,
            sc_instance, 2, pre_unset_contract_calls=[expected])

    def test_new_action_rejected_by_ptg(self):
        ptg1 = self.create_policy_target_group()['policy_target_group']
        ptg2 = self.create_policy_target_group()['policy_target_group']
        ptg3 = self.create_policy_target_group()['policy_target_group']
        simple_rule = self._create_simple_policy_rule()
        simple_rule_2 = self._create_simple_policy_rule()
        prs = self.create_policy_rule_set()['policy_rule_set']
        prs2 = self.create_policy_rule_set(
            policy_rules=[simple_rule['id']])['policy_rule_set']
        prs3 = self.create_policy_rule_set(
            policy_rules=[simple_rule_2['id']])['policy_rule_set']
        self.update_policy_target_group(
            ptg1['id'], provided_policy_rule_sets={prs['id']: ''})
        self.update_policy_target_group(
            ptg2['id'], provided_policy_rule_sets={prs['id']: ''})
        # PTG 3 also consumes a contract
        self.update_policy_target_group(
            ptg3['id'], provided_policy_rule_sets={prs['id']: ''})
        self.update_policy_target_group(
            ptg3['id'], consumed_policy_rule_sets={prs2['id']: ''})

        # Adding a normal rule to PRS works normally
        self.update_policy_rule_set(prs['id'],
                                    policy_rules=[simple_rule['id']])
        redirect = self._create_simple_policy_rule(action_type='redirect')
        self.update_policy_rule_set(
            prs['id'], policy_rules=[simple_rule['id'], redirect['id']],
            expected_res_status=200)

        redirect = self._create_simple_policy_rule(action_type='redirect')
        self.update_policy_rule_set(
            prs3['id'], policy_rules=[simple_rule['id'], redirect['id']],
            expected_res_status=200)
        res = self.update_policy_target_group(
            ptg2['id'], provided_policy_rule_sets={prs['id']: '',
                                                   prs3['id']: ''},
            expected_res_status=400)
        self.assertEqual('PTGAlreadyProvidingRedirectPRS',
                         res['NeutronError']['type'])

        action = self.create_policy_action(
            action_type='redirect')['policy_action']

        self.update_policy_rule_set(
            prs['id'], policy_rules=[simple_rule['id']],
            expected_res_status=200)
        self.update_policy_rule_set(
            prs3['id'], policy_rules=[simple_rule['id']],
            expected_res_status=200)
        # Adding redirect action to the consumed PRS is fine
        self.update_policy_rule(
            simple_rule['id'], policy_actions=[action['id']],
            expected_res_status=200)

    def test_ptg_only_participate_one_prs_when_redirect(self):
        redirect_rule = self._create_simple_policy_rule(action_type='redirect')
        simple_rule = self._create_simple_policy_rule()
        prs_r = self.create_policy_rule_set(
            policy_rules=[redirect_rule['id']])['policy_rule_set']
        prs = self.create_policy_rule_set(
            policy_rules=[simple_rule['id']])['policy_rule_set']

        # Creating PTG with provided redirect and multiple PRS fails
        self.create_policy_target_group(
            provided_policy_rule_sets={prs_r['id']: '', prs['id']: ''},
            consumed_policy_rule_sets={prs['id']: ''},
            expected_res_status=201)

        action = self.create_policy_action(
            action_type='redirect')['policy_action']
        res = self.update_policy_rule(
            simple_rule['id'], policy_actions=[action['id']],
            expected_res_status=400)
        self.assertEqual('PTGAlreadyProvidingRedirectPRS',
                         res['NeutronError']['type'])

    def test_three_tier_sc(self):
        app_db_scs_id = self._create_servicechain_spec(
            node_types=['FIREWALL_TRANSPARENT', 'IDS'], shared=True)
        web_app_scs_id = self._create_servicechain_spec(
            node_types=['FIREWALL_TRANSPARENT', 'LOADBALANCER'])
        internet_web_scs_id = self._create_servicechain_spec(
            node_types=['LOADBALANCER'])
        _, _, app_db_policy_rule_id = self._create_tcp_redirect_rule(
            "20:90", app_db_scs_id)
        _, _, web_app_policy_rule_id = self._create_tcp_redirect_rule(
            "20:90", web_app_scs_id)
        _, _, internet_web_policy_rule_id = self._create_tcp_redirect_rule(
            "20:90", internet_web_scs_id)
        app_db_policy_rule_set = self.create_policy_rule_set(
            policy_rules=[app_db_policy_rule_id])['policy_rule_set']
        web_app_policy_rule_set = self.create_policy_rule_set(
            policy_rules=[web_app_policy_rule_id])['policy_rule_set']
        internet_web_policy_rule_set = self.create_policy_rule_set(
            policy_rules=[internet_web_policy_rule_id])['policy_rule_set']
        # Create DB and APP PTGs on same L2P
        l2p = self.create_l2_policy()['l2_policy']
        db = self.create_policy_target_group(
            l2_policy_id=l2p['id'])['policy_target_group']
        app = self.create_policy_target_group(
            l2_policy_id=l2p['id'])['policy_target_group']

        # Create WEB on a different L2P
        web_l2p = self.create_l2_policy()['l2_policy']
        web = self.create_policy_target_group(
            l2_policy_id=web_l2p['id'])['policy_target_group']

        mgr = self.driver.apic_manager
        mgr.reset_mock()
        self.update_policy_target_group(
            app['id'],
            consumed_policy_rule_sets={app_db_policy_rule_set['id']: ''})
        self.update_policy_target_group(
            db['id'],
            provided_policy_rule_sets={app_db_policy_rule_set['id']: ''})

        sc_instances = self._list_service_chains()
        # We should have one service chain instance created now
        self.assertEqual(len(sc_instances['servicechain_instances']), 1)
        sc_instance = sc_instances['servicechain_instances'][0]
        seen = {sc_instance['id']}

        expected = [
            # Provider EPG provided PRS
            mock.call(db['tenant_id'], db['id'],
                      app_db_policy_rule_set['id'], provider=True,
                      contract_owner=app_db_policy_rule_set['tenant_id'],
                      transaction=mock.ANY),
            # Consumer EPG consumed PRS
            mock.call(app['tenant_id'], app['id'],
                      app_db_policy_rule_set['id'], provider=False,
                      contract_owner=app_db_policy_rule_set['tenant_id'],
                      transaction=mock.ANY)]

        self._verify_chain_set(db, l2p, app_db_policy_rule_set,
                               sc_instance, 2, pre_set_contract_calls=expected)
        mgr.reset_mock()
        self.update_policy_target_group(
            app['id'],
            consumed_policy_rule_sets={app_db_policy_rule_set['id']: ''},
            provided_policy_rule_sets={web_app_policy_rule_set['id']: ''})
        self.update_policy_target_group(
            web['id'],
            consumed_policy_rule_sets={web_app_policy_rule_set['id']: ''})

        expected = [
            # Provider EPG provided PRS
            mock.call(app['tenant_id'], app['id'],
                      web_app_policy_rule_set['id'], provider=True,
                      contract_owner=web_app_policy_rule_set['tenant_id'],
                      transaction=mock.ANY),
            # Consumer EPG consumed PRS
            mock.call(web['tenant_id'], web['id'],
                      web_app_policy_rule_set['id'], provider=False,
                      contract_owner=web_app_policy_rule_set['tenant_id'],
                      transaction=mock.ANY)]

        sc_instances = self._list_service_chains()
        # We should have one service chain instance created now
        self.assertEqual(len(sc_instances['servicechain_instances']), 2)
        sc_instance = [x for x in sc_instances['servicechain_instances']
                       if x['id'] not in seen][0]
        seen.add(sc_instance['id'])

        self._verify_chain_set(app, l2p, web_app_policy_rule_set,
                               sc_instance, 1, pre_set_contract_calls=expected)

        es = self.create_external_segment(
            name='supported', cidr='192.168.0.0/24',
            external_routes=[], expected_res_status=201)['external_segment']

        ep = self.create_external_policy(
            external_segments=[es['id']],
            expected_res_status=201)['external_policy']
        mgr.reset_mock()
        self.update_policy_target_group(
            web['id'],
            consumed_policy_rule_sets={web_app_policy_rule_set['id']: ''},
            provided_policy_rule_sets={internet_web_policy_rule_set['id']: ''})
        self.update_external_policy(
            ep['id'],
            consumed_policy_rule_sets={internet_web_policy_rule_set['id']: ''})
        expected = [
            # Provider EPG provided PRS
            mock.call(web['tenant_id'], web['id'],
                      internet_web_policy_rule_set['id'], provider=True,
                      contract_owner=internet_web_policy_rule_set['tenant_id'],
                      transaction=mock.ANY)]

        sc_instances = self._list_service_chains()
        # We should have one service chain instance created now
        self.assertEqual(len(sc_instances['servicechain_instances']), 3)
        sc_instance = [x for x in sc_instances['servicechain_instances']
                       if x['id'] not in seen][0]
        seen.add(sc_instance['id'])

        self._verify_chain_set(app, web_l2p, internet_web_policy_rule_set,
                               sc_instance, 0, pre_set_contract_calls=expected)

    def test_rule_removed_by_shared(self):
        scs_id = self._create_servicechain_spec(
            node_types=['FIREWALL_TRANSPARENT'], shared=True)
        policy_rule_id = self._create_simple_policy_rule(
            action_type='redirect', shared=True, action_value=scs_id)['id']
        policy_rule_set = self.create_policy_rule_set(
            name="c1", policy_rules=[policy_rule_id],
            shared=True)['policy_rule_set']
        # Create PTGs on same L2P
        l2p = self.create_l2_policy(tenant_id='noadmin')['l2_policy']
        provider = self.create_policy_target_group(
            l2_policy_id=l2p['id'], tenant_id='noadmin')['policy_target_group']

        mgr = self.driver.apic_manager
        # form the chain
        self.update_policy_target_group(
            provider['id'], tenant_id=provider['tenant_id'],
            provided_policy_rule_sets={policy_rule_set['id']: ''})

        mgr.reset_mock()
        self.update_policy_rule_set(policy_rule_set['id'], policy_rules=[],
                                    expected_res_status=200)

    def _verify_chain_set(self, *args, **kwargs):
        pass

    def _verify_chain_unset(self, *args, **kwargs):
        pass

    def _list_service_chains(self):
        sc_instance_list_req = self.new_list_request(
            'servicechain/servicechain_instances')
        res = sc_instance_list_req.get_response(self.ext_api)
        return self.deserialize(self.fmt, res)

    # REVISIT: The following test is being temporarily disabled since the
    # logic in the chain_mapping driver needs to be revisited.
    @unittest2.skip('skipping')
    def test_ha_chain_same_subnet(self):
        session = context.get_admin_context().session
        # Create 2 L3Ps with same pools
        l3p1 = self.create_l3_policy()['l3_policy']
        l3p2 = self.create_l3_policy()['l3_policy']

        # Attach proper L2Ps
        l2p1 = self.create_l2_policy(l3_policy_id=l3p1['id'])['l2_policy']
        l2p2 = self.create_l2_policy(l3_policy_id=l3p2['id'])['l2_policy']

        # Set providers PTGs
        ptg_prov_1 = self.create_policy_target_group(
            l2_policy_id=l2p1['id'])['policy_target_group']
        ptg_prov_2 = self.create_policy_target_group(
            l2_policy_id=l2p2['id'])['policy_target_group']

        # At this point, they  have the same subnet CIDR associated, same will
        # be for Proxy Groups
        ha_spec_id = self._create_servicechain_spec(node_types=['FIREWALL_HA'])

        # Create proper PRS
        policy_rule_id = self._create_simple_policy_rule(
            action_type='redirect', action_value=ha_spec_id)['id']
        policy_rule_set = self.create_policy_rule_set(
            name="c1", policy_rules=[policy_rule_id])['policy_rule_set']

        # Form first and second chain
        ptg_prov_1 = self.update_policy_target_group(
            ptg_prov_1['id'], provided_policy_rule_sets={
                policy_rule_set['id']: ''})['policy_target_group']
        ptg_prov_2 = self.update_policy_target_group(
            ptg_prov_2['id'], provided_policy_rule_sets={
                policy_rule_set['id']: ''})['policy_target_group']

        # One proxy group exists for both, each with 2 service PTs. Put them in
        # HA
        scis = self._list_service_chains()['servicechain_instances']
        self.assertEqual(2, len(scis))
        # Chain 1 targets
        targets_1 = model.get_service_targets(
            session, servicechain_instance_id=scis[0]['id'])
        # Chain 2 targets
        targets_2 = model.get_service_targets(
            session, servicechain_instance_id=scis[1]['id'])

        def _assert_service_targets_and_cluster_them(targets):
            result = {}
            for pt in targets:
                pt = self.show_policy_target(
                    pt.policy_target_id,
                    is_admin_context=True)['policy_target']
                result.setdefault(
                    pt['policy_target_group_id'], []).append(pt)

            # There are 2 PTGs, and 3 PTs each
            self.assertEqual(2, len(result))
            for key in result:
                # Sort by IP address
                result[key].sort(key=lambda x: self._get_object(
                    'ports', x['port_id'],
                    self.api)['port']['fixed_ips'][0]['ip_address'])
                value = result[key]
                self.assertEqual(3, len(value))
                # Set first PT as master of the other twos
                self.update_policy_target(value[1]['id'],
                                          cluster_id=value[0]['id'],
                                          is_admin_context=True)
                self.update_policy_target(value[2]['id'],
                                          cluster_id=value[0]['id'],
                                          is_admin_context=True)
            return result
        # Group chain 1 targets by PTG:
        chain_targets_1 = _assert_service_targets_and_cluster_them(targets_1)
        # Group chain 2 targets by PTG:
        chain_targets_2 = _assert_service_targets_and_cluster_them(targets_2)

        # Verify IPs overlap on provider side
        main_ip = None
        for x in range(3):
            port_1 = self._get_object(
                'ports', chain_targets_1[ptg_prov_1['id']][x]['port_id'],
                self.api)['port']
            port_2 = self._get_object(
                'ports', chain_targets_2[ptg_prov_2['id']][x]['port_id'],
                self.api)['port']
            self.assertEqual(port_1['fixed_ips'][0]['ip_address'],
                             port_2['fixed_ips'][0]['ip_address'])
            if x == 0:
                main_ip = port_1['fixed_ips'][0]['ip_address']

        # Update address ownership on second port
        self.driver.update_ip_owner(
            {'port': chain_targets_1[ptg_prov_1['id']][1]['port_id'],
             'ip_address_v4': main_ip})
        # Same address owned by another port in a different subnet
        self.driver.update_ip_owner(
            {'port': chain_targets_2[ptg_prov_2['id']][1]['port_id'],
             'ip_address_v4': main_ip})

        # There are 2 ownership entries for the same address
        entries = self.driver.ha_ip_handler.session.query(
                    ha_ip_db.HAIPAddressToPortAssocation).all()
        self.assertEqual(2, len(entries))
        self.assertEqual(main_ip, entries[0].ha_ip_address)
        self.assertEqual(main_ip, entries[1].ha_ip_address)


class TestProxyGroup(ApicMappingStitchingPlumberGBPTestCase):

    def _proxy_tenant(self, ptg, admin_proxy):
        return 'admin' if admin_proxy else ptg['tenant_id']

    def _test_proxy_group_same_l2p(self, admin_proxy=False):
        ptg1 = self.create_policy_target_group()['policy_target_group']
        l2p = self.create_l2_policy()['l2_policy']
        proxy = self.create_policy_target_group(
            proxied_group_id=ptg1['id'],
            l2_policy_id=l2p['id'],
            tenant_id=self._proxy_tenant(ptg1, admin_proxy),
            is_admin_context=admin_proxy)['policy_target_group']
        # The used L2P will be ignored, and the proxy will be put on the
        # proxied group's L2P
        self.assertEqual(ptg1['l2_policy_id'], proxy['l2_policy_id'])

    def test_proxy_group_same_l2p(self):
        self._test_proxy_group_same_l2p()

    def test_proxy_group_same_l2p_admin(self):
        self._test_proxy_group_same_l2p(True)

    def _test_l2_proxy_group_subnets(self, admin_proxy=False):
        ptg = self.create_policy_target_group()['policy_target_group']
        proxy = self.create_policy_target_group(
            proxied_group_id=ptg['id'],
            proxy_type='l2', tenant_id=self._proxy_tenant(ptg, admin_proxy),
            is_admin_context=admin_proxy)['policy_target_group']
        ptg = self.show_policy_target_group(ptg['id'])['policy_target_group']
        self.assertEqual(ptg['subnets'], proxy['subnets'])
        self.assertEqual(1, len(proxy['subnets']))

    def test_l2_proxy_group_subnets(self):
        self._test_l2_proxy_group_subnets()

    def test_l2_proxy_group_subnets_admin(self):
        self._test_l2_proxy_group_subnets(True)

    def _test_l3_proxy_group_subnets(self, admin_proxy=False):
        ptg1 = self.create_policy_target_group()['policy_target_group']
        original_subnet = ptg1['subnets'][0]
        proxy = self.create_policy_target_group(
            proxied_group_id=ptg1['id'],
            proxy_type='l3', tenant_id=self._proxy_tenant(ptg1, admin_proxy),
            is_admin_context=admin_proxy)['policy_target_group']

        ptg1 = self.show_policy_target_group(ptg1['id'])['policy_target_group']
        self.assertEqual(set(ptg1['subnets']), set(proxy['subnets']))
        self.assertEqual(2, len(proxy['subnets']))
        proxy['subnets'].remove(original_subnet)
        new_subnet = proxy['subnets'][0]

        # Verify subnet from proxy pool
        l2p = self.show_l2_policy(ptg1['l2_policy_id'])['l2_policy']
        l3p = self.show_l3_policy(l2p['l3_policy_id'])['l3_policy']
        subnet = self._get_object('subnets', new_subnet, self.api)['subnet']
        self.assertTrue(netaddr.IPNetwork(subnet['cidr']) in
                        netaddr.IPNetwork(l3p['proxy_ip_pool']))

        # Attach external segment to create a router interface
        self._mock_external_dict([('supported', '192.168.0.2/24')])
        es = self.create_external_segment(
            name='supported', cidr='192.168.0.0/24',
            external_routes=[{'destination': '0.0.0.0/0',
                              'nexthop': '192.168.0.254'},
                             {'destination': '128.0.0.0/16',
                              'nexthop': None}])['external_segment']
        self.update_l3_policy(l2p['l3_policy_id'],
                              external_segments={es['id']: []},
                              expected_res_status=200)

        # the proxy subnets should have no router
        ports = self._list(
            'ports',
            query_params='device_owner=network:router_interface')['ports']
        self.assertEqual(1, len(ports))

        # this router port is only connected to the original PTG
        self.assertEqual(original_subnet,
                         ports[0]['fixed_ips'][0]['subnet_id'])

        # Verify port address comes from that subnet
        pt = self.create_policy_target(
            policy_target_group_id=proxy['id'],
            tenant_id=proxy['tenant_id'],
            is_admin_context=admin_proxy)['policy_target']
        port = self._get_object('ports', pt['port_id'], self.api)['port']
        self.assertTrue(
            netaddr.IPNetwork(port['fixed_ips'][0]['ip_address']) in
            netaddr.IPNetwork(subnet['cidr']),
            "IP address %s is not part of subnet %s" % (
                port['fixed_ips'][0]['ip_address'], subnet['cidr']))

    def test_l3_proxy_group_subnets(self):
        self._test_l3_proxy_group_subnets()

    def test_l3_proxy_group_subnets_admin(self):
        self._test_l3_proxy_group_subnets(True)

    def test_proxy_shadow_created(self):
        l2p = self.create_l2_policy()['l2_policy']
        ptg = self.create_policy_target_group(
            l2_policy_id=l2p['id'])['policy_target_group']
        self.mgr.reset_mock()
        proxy = self.create_policy_target_group(
            proxied_group_id=ptg['id'])['policy_target_group']
        l2p = self.show_l2_policy(ptg['l2_policy_id'])['l2_policy']
        l3p = self.show_l3_policy(l2p['l3_policy_id'])['l3_policy']
        # Shadow BD created
        self.mgr.ensure_bd_created_on_apic.assert_called_once_with(
            ptg['tenant_id'], 'Shd-' + ptg['id'], ctx_owner=l3p['tenant_id'],
            ctx_name=l3p['id'], allow_broadcast=True, unicast_route=False,
            transaction=mock.ANY, enforce_subnet_check=False)
        # Proxied PTG moved
        expected_calls = [
            # Proxy created on L2P
            mock.call(
                ptg['tenant_id'], proxy['id'], bd_owner=l2p['tenant_id'],
                bd_name=l2p['id']),
            # Proxied moved on shadow BD
            mock.call(
                ptg['tenant_id'], ptg['id'], bd_owner=ptg['tenant_id'],
                bd_name='Shd-' + ptg['id'], transaction=mock.ANY)]
        self._check_call_list(expected_calls,
                              self.mgr.ensure_epg_created.call_args_list)

    def _test_proxy_any_contract(self, admin_proxy=False):
        ptg = self.create_policy_target_group()['policy_target_group']
        proxy = self.create_policy_target_group(
            proxied_group_id=ptg['id'],
            tenant_id=self._proxy_tenant(ptg, admin_proxy),
            is_admin_context=admin_proxy)['policy_target_group']

        self.mgr.create_contract('any-' + ptg['id'], owner='COMMON',
                                 transaction=mock.ANY)
        expected_calls = [
            mock.call(
                ptg['tenant_id'], ptg['id'], 'any-' + ptg['id'], provider=True,
                contract_owner='common'),
            mock.call(
                ptg['tenant_id'], proxy['id'], 'any-' + ptg['id'],
                provider=False, contract_owner='common')]
        self._check_call_list(expected_calls,
                              self.mgr.set_contract_for_epg.call_args_list,
                              check_all=False)

    def test_proxy_any_contract(self):
        self._test_proxy_any_contract()

    def test_proxy_any_contract_admin(self):
        self._test_proxy_any_contract(True)

    def _test_proxy_shadow_deleted(self, admin_proxy=False):
        ptg1 = self.create_policy_target_group()['policy_target_group']
        original_subnet = ptg1['subnets'][0]
        proxy = self.create_policy_target_group(
            proxied_group_id=ptg1['id'],
            tenant_id=self._proxy_tenant(ptg1, admin_proxy),
            is_admin_context=admin_proxy)['policy_target_group']

        self.mgr.reset_mock()
        self.delete_policy_target_group(proxy['id'], expected_res_status=204)
        proxy['subnets'].remove(original_subnet)
        # Proxied PTG moved back
        self.mgr.ensure_epg_created.assert_called_once_with(
            ptg1['tenant_id'], ptg1['id'], bd_owner=ptg1['tenant_id'],
            bd_name=ptg1['l2_policy_id'])
        # Shadow BD deleted
        self.mgr.delete_bd_on_apic.assert_called_once_with(
            ptg1['tenant_id'], 'Shd-' + ptg1['id'])
        # Verify Jump subnet deleted
        self.assertEqual(1, len(proxy['subnets']))
        self._get_object('subnets', proxy['subnets'][0], self.api,
                         expected_res_status=404)

    def _test_get_gbp_details(self, admin_proxy=False, async=False):
        def request_wrapper(*args, **kwargs):
            kwargs['timestamp'] = 0
            kwargs['request_id'] = 'some_id'
            result = self.driver.request_endpoint_details(*args,
                                                          request=kwargs)
            if result:
                return result.get('gbp_details')

        gbp_details = {False: self.driver.get_gbp_details,
                       True: request_wrapper}
        ptg = self.create_policy_target_group(
            name="ptg1")['policy_target_group']
        pt1 = self.create_policy_target(
            policy_target_group_id=ptg['id'])['policy_target']
        data = {'allowed_address_pairs':
                [{'ip_address': '170.166.0.1'},
                 {'ip_address': '170.166.0.2'}]}
        # Create EP with bound port
        port = self.driver._update_port(context.get_admin_context(),
                                        pt1['port_id'], data)
        self.assertEqual(['170.166.0.1', '170.166.0.2'],
                         [x['ip_address'] for x in
                          port['allowed_address_pairs']])
        self._bind_port_to_host(pt1['port_id'], 'h1')
        pt2 = self.create_policy_target(
            policy_target_group_id=ptg['id'])['policy_target']
        self._bind_port_to_host(pt2['port_id'], 'h2')

        # Create proxy group
        proxy = self.create_policy_target_group(
            proxied_group_id=ptg['id'],
            tenant_id=self._proxy_tenant(ptg, admin_proxy),
            is_admin_context=admin_proxy)['policy_target_group']
        proxy_gw = self.create_policy_target(
            policy_target_group_id=proxy['id'],
            proxy_gateway=True,
            tenant_id=proxy['tenant_id'],
            is_admin_context=admin_proxy)['policy_target']
        self._bind_port_to_host(proxy_gw['port_id'], 'h2')
        # Create a PT in the same cluster
        proxy_gw_failover = self.create_policy_target(
            policy_target_group_id=proxy['id'],
            cluster_id=proxy_gw['id'],
            tenant_id=proxy['tenant_id'],
            is_admin_context=admin_proxy)['policy_target']
        self._bind_port_to_host(proxy_gw_failover['port_id'], 'h2')
        self._mock_external_dict([('supported', '192.168.0.2/24')])
        es = self.create_external_segment(
            name='supported', cidr='192.168.0.0/24',
            external_routes=[], expected_res_status=201)['external_segment']
        self.create_external_policy(external_segments=[es['id']],
                                    expected_res_status=201)

        l2p = self.show_l2_policy(ptg['l2_policy_id'])['l2_policy']
        self.update_l3_policy(l2p['l3_policy_id'],
                              external_segments={es['id']: []},
                              expected_res_status=200)

        self._bind_port_to_host(pt2['port_id'], 'h2')
        master_port = self._get_object('ports', proxy_gw['port_id'],
                                       self.api)['port']

        def echo(name):
            return name
        self.mgr.apic.fvTenant.name = echo
        mapping = gbp_details[async](
            context.get_admin_context(),
            device='tap%s' % proxy_gw['port_id'], host='h2')

        # Verify extra addresses
        ips = self._get_pts_addresses([pt1, pt2])
        self.assertEqual(set(ips + ['170.166.0.1', '170.166.0.2']),
                         set(mapping['extra_ips']))
        self.assertEqual(ptg['tenant_id'], mapping['ptg_tenant'])
        self.assertEqual(1, len(mapping['ip_mapping']))
        # No SNAT subnet
        self.assertEqual(0, len(mapping['host_snat_ips']))

        group_default_gw = self.create_policy_target(
            policy_target_group_id=ptg['id'],
            group_default_gateway=True,
            tenant_id=ptg['tenant_id'])['policy_target']
        self._bind_port_to_host(pt2['port_id'], 'h2')
        mapping = gbp_details[async](
            context.get_admin_context(),
            device='tap%s' % group_default_gw['port_id'], host='h2')
        self.assertTrue(mapping['promiscuous_mode'])

        # No extra IPs for the failover since it doesn't own the master IP
        mapping = gbp_details[async](
            context.get_admin_context(),
            device='tap%s' % proxy_gw_failover['port_id'], host='h2')
        self.assertEqual(0, len(mapping['extra_ips'] or []))
        self.assertEqual(
            [{'mac_address': master_port['mac_address'],
             'ip_address': master_port['fixed_ips'][0]['ip_address']}],
            mapping['allowed_address_pairs'])
        # Set the port ownership and verify that extra_ips is correctly set
        ips = self._get_pts_addresses([pt1, pt2, group_default_gw])
        for x in master_port['fixed_ips']:
            self.driver.ha_ip_handler.set_port_id_for_ha_ipaddress(
                proxy_gw_failover['port_id'], x['ip_address'])

        mapping = gbp_details[async](
            context.get_admin_context(),
            device='tap%s' % proxy_gw_failover['port_id'], host='h2')
        self.assertEqual(
            set(ips + ['170.166.0.1', '170.166.0.2']),
            set(mapping['extra_details'][master_port['mac_address']][
                    'extra_ips']))
        self.assertEqual(
            [{'mac_address': master_port['mac_address'],
             'ip_address': master_port['fixed_ips'][0]['ip_address'],
             'active': True}],
            mapping['allowed_address_pairs'])

    def test_get_gbp_details(self):
        self._test_get_gbp_details()

    def test_get_gbp_details_admin(self):
        self._test_get_gbp_details(True)

    def test_get_gbp_details_async(self):
        self._test_get_gbp_details(False, True)

    def test_get_gbp_details_admin_async(self):
        self._test_get_gbp_details(True, True)

    def test_cluster_promiscuous_mode(self):
        ptg = self.create_policy_target_group(
            name="ptg1")['policy_target_group']
        # Create proxy group
        self.create_policy_target_group(proxied_group_id=ptg['id'])
        group_gw = self.create_policy_target(
            policy_target_group_id=ptg['id'],
            group_default_gateway=True)['policy_target']
        # Create a PT in the same cluster
        group_gw_failover = self.create_policy_target(
            policy_target_group_id=ptg['id'],
            cluster_id=group_gw['id'])['policy_target']
        mapping = self.driver.get_gbp_details(
            context.get_admin_context(),
            device='tap%s' % group_gw_failover['port_id'], host='h2')
        self.assertTrue(mapping['promiscuous_mode'])

    def _test_end_chain_notified(self, admin_proxy=False):
        self.driver.notifier = mock.Mock()
        ptg = self.create_policy_target_group(
            name="ptg1")['policy_target_group']
        proxy1 = self.create_policy_target_group(
            proxied_group_id=ptg['id'],
            tenant_id=self._proxy_tenant(ptg, admin_proxy),
            is_admin_context=admin_proxy)['policy_target_group']
        proxy2 = self.create_policy_target_group(
            proxied_group_id=proxy1['id'],
            tenant_id=self._proxy_tenant(ptg, admin_proxy),
            is_admin_context=admin_proxy)['policy_target_group']
        proxy_gw = self.create_policy_target(
            policy_target_group_id=proxy2['id'],
            proxy_gateway=True, tenant_id=proxy2['tenant_id'],
            is_admin_context=admin_proxy)['policy_target']
        self._bind_port_to_host(proxy_gw['port_id'], 'h2')
        self.driver.notifier.reset_mock()

        # Create a PT on the starting PTG, and verify that proxy_gw is
        # notified
        self.create_policy_target(policy_target_group_id=ptg['id'])
        self.assertEqual(1, self.driver.notifier.port_update.call_count)
        self.assertEqual(
            proxy_gw['port_id'],
            self.driver.notifier.port_update.call_args[0][1]['id'])

    def _test_end_chain_notified_cluster_id(self, admin_proxy=False):
        self.driver.notifier = mock.Mock()
        ptg = self.create_policy_target_group(
            name="ptg1")['policy_target_group']
        proxy1 = self.create_policy_target_group(
            proxied_group_id=ptg['id'],
            tenant_id=self._proxy_tenant(ptg, admin_proxy),
            is_admin_context=admin_proxy)['policy_target_group']
        proxy2 = self.create_policy_target_group(
            proxied_group_id=proxy1['id'],
            tenant_id=self._proxy_tenant(ptg, admin_proxy),
            is_admin_context=admin_proxy)['policy_target_group']
        proxy_gw_master = self.create_policy_target(
            policy_target_group_id=proxy2['id'],
            proxy_gateway=True, tenant_id=proxy2['tenant_id'],
            is_admin_context=admin_proxy)['policy_target']
        # The following is not proxy gateway, but is part of a proxy_gw cluster
        proxy_gw = self.create_policy_target(
            policy_target_group_id=proxy2['id'],
            cluster_id=proxy_gw_master['id'], tenant_id=proxy2['tenant_id'],
            is_admin_context=admin_proxy)['policy_target']

        self._bind_port_to_host(proxy_gw['port_id'], 'h2')
        self.driver.notifier.reset_mock()

        # Create a PT on the starting PTG, and verify that proxy_gw is
        # notified
        self.create_policy_target(policy_target_group_id=ptg['id'])
        self.assertEqual(1, self.driver.notifier.port_update.call_count)
        self.assertEqual(
            proxy_gw['port_id'],
            self.driver.notifier.port_update.call_args[0][1]['id'])

    def test_end_chain_notified(self):
        self._test_end_chain_notified()

    def test_end_chain_notified_admin(self):
        self._test_end_chain_notified(True)

    def test_proxy_group_right_tenant(self):
        l2p = self.create_l2_policy(tenant_id='non-admin')['l2_policy']
        ptg = self.create_policy_target_group(
            l2_policy_id=l2p['id'],
            tenant_id='non-admin')['policy_target_group']
        self.mgr.reset_mock()
        proxy = self.create_policy_target_group(
            proxied_group_id=ptg['id'], is_admin_context=True,
            tenant_id='admin-tenant')['policy_target_group']
        l3p = self.show_l3_policy(l2p['l3_policy_id'],
                                  tenant_id='non-admin')['l3_policy']

        # Shadow BD created in non-admin tenant
        self.mgr.ensure_bd_created_on_apic.assert_called_once_with(
            'non-admin', 'Shd-' + ptg['id'], ctx_owner=l3p['tenant_id'],
            ctx_name=l3p['id'], allow_broadcast=True, unicast_route=False,
            transaction=mock.ANY, enforce_subnet_check=False)
        # Proxied PTG moved
        expected_calls = [
            # Proxy created on L2P
            mock.call(
                'non-admin', proxy['id'], bd_owner=l2p['tenant_id'],
                bd_name=l2p['id']),
            # Proxied moved on shadow BD
            mock.call(
                'non-admin', ptg['id'], bd_owner=ptg['tenant_id'],
                bd_name='Shd-' + ptg['id'], transaction=mock.ANY)]
        self._check_call_list(expected_calls,
                              self.mgr.ensure_epg_created.call_args_list)

    def test_prs_sync_with_proxy(self, admin_proxy=False):
        rule = self._create_ssh_allow_rule()
        policy_rule_set_1 = self.create_policy_rule_set(
            name="c1", policy_rules=[rule['id']])['policy_rule_set']
        policy_rule_set_2 = self.create_policy_rule_set(
            name="c2", policy_rules=[rule['id']])['policy_rule_set']
        ptg = self.create_policy_target_group(
            provided_policy_rule_sets={policy_rule_set_1['id']: ''},
            consumed_policy_rule_sets={policy_rule_set_2['id']: ''})[
                'policy_target_group']
        # Sync on proxy creation
        proxy1 = self.create_policy_target_group(
            proxied_group_id=ptg['id'],
            tenant_id=self._proxy_tenant(ptg, admin_proxy),
            is_admin_context=admin_proxy)['policy_target_group']

        self.assertEqual([policy_rule_set_1['id']],
                         proxy1['provided_policy_rule_sets'])
        self.assertEqual([policy_rule_set_2['id']],
                         proxy1['consumed_policy_rule_sets'])
        proxy2 = self.create_policy_target_group(
            proxied_group_id=proxy1['id'],
            tenant_id=self._proxy_tenant(ptg, admin_proxy),
            is_admin_context=admin_proxy)['policy_target_group']
        self.assertEqual([policy_rule_set_1['id']],
                         proxy2['provided_policy_rule_sets'])
        self.assertEqual([policy_rule_set_2['id']],
                         proxy2['consumed_policy_rule_sets'])

        # Sync on original update
        self.update_policy_target_group(
            ptg['id'], provided_policy_rule_sets={})
        proxy1 = self.show_policy_target_group(
            proxy1['id'])['policy_target_group']
        self.assertEqual([],
                         proxy1['provided_policy_rule_sets'])
        self.assertEqual([policy_rule_set_2['id']],
                         proxy1['consumed_policy_rule_sets'])
        proxy2 = self.show_policy_target_group(
            proxy2['id'])['policy_target_group']
        self.assertEqual([],
                         proxy1['provided_policy_rule_sets'])
        self.assertEqual([policy_rule_set_2['id']],
                         proxy2['consumed_policy_rule_sets'])

    def test_l3_proxy_subnets_delete(self):
        l2p = self.create_l2_policy(tenant_id='non-admin')['l2_policy']
        ptg = self.create_policy_target_group(
            l2_policy_id=l2p['id'],
            tenant_id='non-admin')['policy_target_group']
        main_subnet = set(ptg['subnets'])
        self.mgr.reset_mock()
        proxy = self.create_policy_target_group(
            proxied_group_id=ptg['id'], is_admin_context=True,
            tenant_id='admin-tenant')['policy_target_group']
        subnets = set(proxy['subnets'])
        added = subnets - main_subnet
        self.delete_policy_target_group(proxy['id'], is_admin_context=True,
                                        expected_res_status=204)
        # Subnet doesn't exist anymore
        self._get_object('subnets', list(added)[0], self.api,
                         expected_res_status=404)


class TestApicChainsAdminOwner(TestApicChains):

    def setUp(self, **kwargs):
        mock.patch('gbpservice.neutron.services.grouppolicy.drivers.'
                   'chain_mapping.ChainMappingDriver.'
                   'chain_tenant_keystone_client').start()
        res = mock.patch('gbpservice.neutron.services.grouppolicy.drivers.'
                         'chain_mapping.ChainMappingDriver.'
                         'chain_tenant_id').start()
        res.return_value = test_rmd.CHAIN_TENANT_ID
        super(TestApicChainsAdminOwner, self).setUp(**kwargs)

    def _assert_proper_chain_instance(self, sc_instance, provider_ptg_id,
                                      consumer_ptg_id, scs_id_list,
                                      classifier_id=None):
        self.assertEqual(sc_instance['provider_ptg_id'], provider_ptg_id)
        self.assertEqual(sc_instance['consumer_ptg_id'], 'N/A')
        self.assertEqual(sc_instance['tenant_id'], test_rmd.CHAIN_TENANT_ID)
        self.assertEqual(scs_id_list, sc_instance['servicechain_specs'])
        if classifier_id:
            self.assertEqual(sc_instance['classifier_id'], classifier_id)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import commands
import logging
import six
import sys

from libs import config_libs
from libs import utils_libs
from libs import verify_libs


def main():

    # Run the Testcases:
    test = test_gbp_l3p_neg()
    if test.test_gbp_l3p_neg_1() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_L3P_NEG_1')
    if test.test_gbp_l3p_neg_2() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_L3P_NEG_2')
    if test.test_gbp_l3p_neg_3() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_L3P_NEG_3')
    if test.test_gbp_l3p_neg_4() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_L3P_NEG_4')
    if test.test_gbp_l3p_neg_5() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_L3P_NEG_5')
    test.cleanup()
    utils_libs.report_results('test_gbp_l3p_neg', 'test_results.txt')
    sys.exit(1)


class test_gbp_l3p_neg(object):

    # Initialize logging
    logging.basicConfig(
        format='%(asctime)s [%(levelname)s] %(name)s - %(message)s',
        level=logging.WARNING)
    _log = logging.getLogger(__name__)
    cmd = 'rm /tmp/test_gbp_l3p_neg.log'
    commands.getoutput(cmd)
    hdlr = logging.FileHandler('/tmp/test_gbp_l3p_neg.log')
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    hdlr.setFormatter(formatter)
    _log.addHandler(hdlr)
    _log.setLevel(logging.INFO)
    _log.setLevel(logging.DEBUG)

    def __init__(self):
        """
        Init def
        """
        self._log.info("\n## START OF GBP L3_POLICY NEGATIVE TESTSUITE\n")
        self.gbpcfg = config_libs.Gbp_Config()
        self.gbpverify = verify_libs.Gbp_Verify()
        self.l3p_name = 'demo_l3p'

    def cleanup(self, tc_name=''):
        if tc_name != '':
            self._log.info('%s: FAILED' % (tc_name))
        for obj in ['group', 'l2p', 'l3p']:
            self.gbpcfg.gbp_del_all_anyobj(obj)

    def test_gbp_l3p_neg_1(self):

        self._log.info(
            "\n#############################################\n"
            "TESTCASE_GBP_L3P_NEG_1: TO CREATE/VERIFY L3POLICY "
            "with INVALID IP-POOL\n"
            "TEST_STEPS::\n"
            "Create L3Policy Object with Invalid IP-Pool\n"
            "Invalid IP-Pools: x.y.0.0/24, 0.0.0.0/0,255.255.255.255/32,"
            "0.2323.0.0/24\n"
            "Verify the create FAILs and config rolls back\n"
            "############################################\n")

        # Testcase work-flow starts
        count = 0
        invalid_pools = [
            'x.y.0.0/24',
            '0.2323.0.0/24',
            '0.0.0.0/0',
            '255.255.255.255/32']
        for pool in invalid_pools:
            self._log.info(
                "\n## Step 1A: Create L3Policy with Invalid IP-Pool = %s ##" %
                (pool))
            if self.gbpcfg.gbp_policy_cfg_all(
                    1, 'l3p', self.l3p_name, ip_pool=pool) != 0:
                self._log.info(
                    "# Step 1A: Create L3Policy with Invalid IP-Pool %s did "
                    "NOT fail" %
                    (pool))
            self._log.info('# Step 1A: Verify L3Policy did NOT get created')
            if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                    1, 'l3p', self.l3p_name) != 0:
                self._log.info(
                    "# Step 1A: L3Policy did NOT fail to create even with "
                    "Invalid IP-Pool %s" %
                    (pool))
                count += 1
        if count > 0:
            return 0
        else:
            self._log.info("\nTESTCASE_GBP_L3P_NEG_1: PASSED")
            return 1

    def test_gbp_l3p_neg_2(self):

        self._log.info(
            "\n############################################\n"
            "TESTCASE_GBP_L3P_NEG_2: TO CREATE/VERIFY L3POLICY with INVALID "
            "SUBNET-PREF-LENGTH\n"
            "TEST_STEPS::\n"
            "Create L3Policy Object with Invalid Subnet-Prefix-Length\n"
            "Invalid Subnet-Prefix-Lengths: 33,'AB','32'\n"
            "Verify the create FAILs and config rolls back\n"
            "############################################\n")

        # Testcase work-flow starts
        cnt = 0
        invalid_prefix_length = ['33', 'AB', '32']
        for prefix in invalid_prefix_length:
            self._log.info(
                "\n## Step 1A: Create L3Policy with Invalid "
                "Prefix-lenght = %s ##" %
                (prefix))
            if self.gbpcfg.gbp_policy_cfg_all(
                    1, 'l3p', self.l3p_name, subnet_prefix_length=prefix) != 0:
                self._log.info(
                    "# Step 1A: Create L3Policy with Invalid IP-Pool %s "
                    "did NOT fail" %
                    (prefix))
            self._log.info('# Step 1A: Verify L3Policy did NOT get created')
            if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                    1, 'l3p', self.l3p_name) != 0:
                self._log.info(
                    "# Step 1A: L3Policy did NOT fail to create even with "
                    "Invalid IP-Pool %s" %
                    (prefix))
                cnt += 1
        if cnt > 0:
            return 0
        else:
            self._log.info("\nTESTCASE_GBP_L3P_NEG_2: PASSED")
            return 1

    def test_gbp_l3p_neg_3(self):

        self._log.info(
            "\n############################################\n"
            "TESTCASE_GBP_L3P_NEG_3: TO CREATE/VERIFY L3POLICY with mix "
            "of VALID & INVALID ATTRs\n"
            "TEST_STEPS::\n"
            "Create L3Policy with a mix of Valid IP-Pool and Invalid "
            "Subnet-Prefix-Length & Vice-versa\n"
            "Invalid IP-Pool: x.y.0.0/24,Valid Subnet-Pref-Len: 30\n"
            "Valid IP-Pool: 20.20.20.0/24, Invalid Subnet-Pref-Len: 32\n"
            "Verify the create FAILs and config rolls back\n"
            "############################################\n")

        # Testcase work-flow starts
        mix_attr = {'x.y.0.0/24': '30', '20.20.20.0/24': '32'}
        _pass = 0
        for ip, pref in six.iteritems(mix_attr):
            self._log.info(
                "\n## Step 1A: Create L3Policy with IP-Pool = %s & "
                "Subnet-Pref-Len = %s ##" %
                (ip, pref))
            if self.gbpcfg.gbp_policy_cfg_all(
                    1,
                    'l3p',
                    self.l3p_name,
                    ip_pool=ip,
                    subnet_prefix_length=pref) != 0:
                self._log.info(
                    "# Step 1A: Create L3Policy with mix of valid and "
                    "invalid did NOT fail")
            self._log.info('# Step 1A: Verify L3Policy did NOT get created')
            if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                    1, 'l3p', self.l3p_name) != 0:
                self._log.info(
                    "# Step 1A: L3Policy did NOT fail to create even with "
                    "mix of Valid and Invalid attrs %s")
                _pass += 1
        if _pass > 0:
            return 0
        else:
            self._log.info("\nTESTCASE_GBP_L3P_NEG_3: PASSED")
            return 1

    def test_gbp_l3p_neg_4(self):

        self._log.info(
            "\n#################################################\n"
            "TESTCASE_GBP_L3P_NEG_4: TO UPDATE/VERIFY L3POLICY with "
            "INVALID ATTRs\n"
            "TEST_STEPS::\n"
            "Create a L3Policy with default attr values\n"
            "Update the L3Policy with Invalid Subnet-Prefix-Length\n"
            "Update the L3Policy with Valid IP-Pool, should fail as "
            "ip-pool is Immutable attr\n"
            "Verify the update fails and config roll backs to original "
            "values of the L3Policy\n"
            "###############################################\n")

        # Testcase work-flow starts
        self._log.info('\n## Step 1: Create a L3P with default attribute ##\n')
        l3p_uuid = self.gbpcfg.gbp_policy_cfg_all(1, 'l3p', self.l3p_name)
        if l3p_uuid == 0:
            self._log.info("\n## Step 1: Create L3Policy == Failed")
            return 0
        if self.gbpcfg.gbp_policy_cfg_all(
                2, 'l3p', l3p_uuid, subnet_prefix_length='32') != 0:
            self._log.info(
                "\n## Step 2: Updating L3Policy's Subnet-Prefix-Length "
                "with Invalid Value=32 did NOT fail")
            return 0
        if self.gbpcfg.gbp_policy_cfg_all(
                2, 'l3p', l3p_uuid, ip_pool='20.20.0.0/24') != 0:
            self._log.info(
                "\n## Step 3: Updating L3Policy's Immutable attr IP-Pool "
                "did NOT fail")
            return 0
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                1,
                'l3p',
                l3p_uuid,
                id=l3p_uuid,
                name=self.l3p_name,
                ip_pool='10.0.0.0/8',
                subnet_prefix_length='24') == 0:
            self._log.info(
                "\n## Step 4: L3Policy config did NOT roll back to original "
                "default values")
            return 0
        self.gbpcfg.gbp_policy_cfg_all(
            0, 'l3p', l3p_uuid)  # clean-up before next testcase
        self._log.info("\nTESTCASE_GBP_L3P_NEG_4: PASSED")
        return 1

    def test_gbp_l3p_neg_5(self):

        self._log.info(
            "\n#################################################\n"
            "TESTCASE_GBP_L3P_NEG_5: TO CREATE/UPDATE L3POLICY with "
            "SUBNET-PREF-LENGTH GREATER than IP-POOL's MASK-LENGTH\n"
            "TEST_STEPS::\n"
            "Create a L3Policy with non-default attr, "
            "subnet-pref-length > mask-length of pool\n"
            "Verify the above L3Policy creation fails\n"
            "Create a L3Policy with default attrs\n"
            "Update the L3Policy's subnet-pref-length such that "
            "subnet-pref-length > mask-length of pool\n"
            "Verify the update fails and L3Policy attrs persists with "
            "default values\n"
            "##################################################\n")

        # Testcase work-flow starts
        self._log.info('\n## Step 1: Create a L3P with default attribute ##\n')
        l3p_uuid = self.gbpcfg.gbp_policy_cfg_all(1, 'l3p', self.l3p_name)
        if l3p_uuid == 0:
            self._log.info("\n## Step 1: Create L3Policy == Failed")
            return 0
        if self.gbpcfg.gbp_policy_cfg_all(
                2, 'l3p', l3p_uuid, subnet_prefix_length='4') != 0:
            self._log.info(
                "\n## Step 2: Updating L3Policy's "
                "Subnet-Prefix-Length > default Mask-length(8) did NOT fail")
            return 0
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                1,
                'l3p',
                l3p_uuid,
                id=l3p_uuid,
                name=self.l3p_name,
                ip_pool='10.0.0.0/8',
                subnet_prefix_length='24') == 0:
            self._log.info(
                "\n## Step 3: L3Policy config did NOT roll back "
                "to original default values")
            return 0
        if self.gbpcfg.gbp_policy_cfg_all(
                1,
                'l3p',
                'new_l3p',
                ip_pool='20.20.20.0/24',
                subnet_prefix_length='16') != 0:
            self._log.info(
                "\n## Step 4: Creating L3Policy with "
                "Subnet-Prefix-Length > Mask-Length(24) did NOT fail")
            return 0
        self._log.info("\nTESTCASE_GBP_L3P_NEG_5: PASSED")
        return 1

if __name__ == '__main__':
    main()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ast
from collections import defaultdict
from gbpservice._i18n import _LE
from gbpservice._i18n import _LW

from gbpservice.nfp.common import constants as nfp_constants
from gbpservice.nfp.common import data_formatter as df
from gbpservice.nfp.common import exceptions
from gbpservice.nfp.core import executor as nfp_executor
from gbpservice.nfp.core import log as nfp_logging
from gbpservice.nfp.lib import nfp_context_manager as nfp_ctx_mgr
from gbpservice.nfp.orchestrator.coal.networking import (
    nfp_gbp_network_driver
)
from gbpservice.nfp.orchestrator.coal.networking import (
    nfp_neutron_network_driver
)
from gbpservice.nfp.orchestrator.openstack import openstack_driver

LOG = nfp_logging.getLogger(__name__)


def _set_network_handler(f):
    def wrapped(self, *args, **kwargs):
        if type(args[0]) == dict:
            device_data = args[0]
        else:
            device_data = args[1]
        if device_data.get('service_details'):
            network_mode = device_data['service_details'].get('network_mode')
            if network_mode:
                kwargs['network_handler'] = self.network_handlers[network_mode]
        return f(self, *args, **kwargs)
    return wrapped


class OrchestrationDriver(object):
    """Generic Driver class for orchestration of virtual appliances

    Launches the VM with all the management and data ports and a new VM
    is launched for each Network Service Instance
    """

    def __init__(self, config, supports_device_sharing=False,
                 supports_hotplug=False, max_interfaces=8):
        self.service_vendor = 'general'
        self.supports_device_sharing = supports_device_sharing
        self.supports_hotplug = supports_hotplug
        self.maximum_interfaces = max_interfaces
        self.identity_handler = openstack_driver.KeystoneClient(config)
        self.compute_handler_nova = openstack_driver.NovaClient(config)
        self.network_handlers = {
            nfp_constants.GBP_MODE:
                nfp_gbp_network_driver.NFPGBPNetworkDriver(config),
            nfp_constants.NEUTRON_MODE:
                nfp_neutron_network_driver.NFPNeutronNetworkDriver(config)
        }
        self.config = config

    def _get_admin_tenant_id(self, token=None):
        with nfp_ctx_mgr.KeystoneContextManager as kcm:
            if not token:
                token = kcm.retry(
                    self.identity_handler.get_admin_token, tries=3)
            admin_tenant_name = (
                self.config.nfp_keystone_authtoken.admin_tenant_name)
            admin_tenant_id = kcm.retry(self.identity_handler.get_tenant_id,
                                        token,
                                        admin_tenant_name, tries=3)
            return admin_tenant_id

    def _get_token(self, device_data_token):

        with nfp_ctx_mgr.KeystoneContextManager as kcm:
            token = (device_data_token
                     if device_data_token
                     else kcm.retry(
                         self.identity_handler.get_admin_token, tries=3))
            return token

    def _is_device_sharing_supported(self):
        return False

    def _create_management_interface(self, token, admin_tenant_id,
                                     device_data, network_handler):

        name = nfp_constants.MANAGEMENT_INTERFACE_NAME
        mgmt_interface = network_handler.create_port(
            token,
            admin_tenant_id,
            device_data['management_network_info']['id'],
            name=name)

        return {'id': mgmt_interface['id'],
                'port_id': mgmt_interface['port_id'],
                'port_model': (nfp_constants.GBP_PORT
                               if device_data['service_details'][
                                   'network_mode'] ==
                               nfp_constants.GBP_MODE
                               else nfp_constants.NEUTRON_PORT),
                'port_classification': nfp_constants.MANAGEMENT,
                'port_role': None}

    def _delete_interface(self, device_data, interface,
                          network_handler=None):
        token = self._get_token(device_data.get('token'))
        if not token:
            return None

        network_handler.delete_port(token, interface['id'])

    def _get_interfaces_for_device_create(self, token, admin_tenant_id,
                                          network_handler, device_data):
        try:
            mgmt_interface = self._create_management_interface(
                token,
                admin_tenant_id,
                device_data,
                network_handler)
            device_data['interfaces'] = [mgmt_interface]
        except Exception as e:
            LOG.exception(_LE('Failed to get interfaces for device creation.'
                              'Error: %(error)s'), {'error': e})

    def _delete_interfaces(self, device_data, interfaces,
                           network_handler=None):
        for interface in interfaces:
            self._delete_interface(
                device_data, interface,
                network_handler=network_handler)

    def _verify_provider_metadata(self, image_name, metadata):
        provider_metadata = {}
        try:
            for attr in metadata:
                if attr in nfp_constants.METADATA_SUPPORTED_ATTRIBUTES:
                    provider_metadata[attr] = ast.literal_eval(metadata[attr])
        except Exception as e:
            LOG.error(_LE('Wrong metadata: %(metadata)s provided for '
                          'image name: %(image_name)s. Error: %(error)s'),
                      {'image_name': image_name, 'metadata': metadata,
                       'error': e})
            return None
        return provider_metadata

    def _get_provider_metadata(self, device_data, image_name):
        token = self._get_token(device_data.get('token'))
        if not token:
            return None
        with nfp_ctx_mgr.NovaContextManager as ncm:
            metadata = ncm.retry(self.compute_handler_nova.get_image_metadata,
                                 token,
                                 self._get_admin_tenant_id(token=token),
                                 image_name)
        provider_metadata = self._verify_provider_metadata(
            image_name, metadata)
        if not provider_metadata:
            return {}
        return provider_metadata

    def _get_provider_metadata_fast(self, token,
                                    admin_tenant_id, image_name, device_data):
        with nfp_ctx_mgr.NovaContextManager as ncm:
            metadata = ncm.retry(self.compute_handler_nova.get_image_metadata,
                                 token,
                                 admin_tenant_id,
                                 image_name)
        provider_metadata = self._verify_provider_metadata(
            image_name, metadata)
        if not provider_metadata:
            return {}
        return provider_metadata

    def _update_self_with_provider_metadata(self, provider_metadata, attr):
        attr_value = getattr(self, attr)
        if attr in provider_metadata:
            setattr(self, attr, provider_metadata[attr])
        else:
            LOG.debug("Provider metadata specified in image, doesn't contains "
                      "%s value, proceeding with default value "
                      "%s", (attr, attr_value))

    def _update_provider_metadata(self, device_data, token=None):
        provider_metadata = {}
        try:
            image_name = self._get_image_name(device_data)
            provider_metadata = self._get_provider_metadata(device_data,
                                                            image_name)
            LOG.debug("Provider metadata, specified in image: %s",
                      provider_metadata)
            if provider_metadata:
                self._update_self_with_provider_metadata(
                    provider_metadata,
                    nfp_constants.MAXIMUM_INTERFACES)
                self._update_self_with_provider_metadata(
                    provider_metadata,
                    nfp_constants.SUPPORTS_HOTPLUG)
            else:
                LOG.debug("No provider metadata specified in image,"
                          " proceeding with default values")
        except Exception:
            LOG.error(_LE("Error while getting metadata for image name:"
                          "%(image_name)s, proceeding with default values"),
                      {'image_name': image_name})
        return provider_metadata

    def _update_provider_metadata_fast(self, token, admin_tenant_id,
                                       image_name, device_data):
        provider_metadata = None
        try:
            provider_metadata = self._get_provider_metadata_fast(
                token, admin_tenant_id, image_name, device_data)
            LOG.debug("Provider metadata, specified in image: %s",
                      provider_metadata)
            if provider_metadata:
                self._update_self_with_provider_metadata(
                    provider_metadata,
                    nfp_constants.MAXIMUM_INTERFACES)
                self._update_self_with_provider_metadata(
                    provider_metadata,
                    nfp_constants.SUPPORTS_HOTPLUG)
            else:
                LOG.debug("No provider metadata specified in image,"
                          " proceeding with default values")
        except Exception:
            LOG.error(_LE("Error while getting metadata for image name: "
                          "%(image_name)s, proceeding with default values"),
                      {'image_name': image_name})
        return provider_metadata

    def _get_image_name(self, device_data):
        if device_data['service_details'].get('image_name'):
            image_name = device_data['service_details']['image_name']
        else:
            LOG.debug("No image name provided in service profile's "
                      "service flavor field, image will be selected "
                      "based on service vendor's name : %s",
                      (device_data['service_details']['service_vendor']))
            image_name = device_data['service_details']['service_vendor']
            image_name = '%s' % image_name.lower()
            device_data['service_details']['image_name'] = image_name
        return image_name

    def _get_service_type(self, token, service_profile_id, network_handler):
        service_profile = network_handler.get_service_profile(
            token, service_profile_id)
        return service_profile['service_type']

    def _get_device_service_types_map(self, token, devices, network_handler):
        device_service_types_map = defaultdict(set)
        for device in devices:
            for network_function in device['network_functions']:
                service_type = self._get_service_type(
                    token,
                    network_function['service_profile_id'],
                    network_handler)
                device_service_types_map[device['id']].add(service_type)
        return device_service_types_map

    def get_image_id(self, nova, token, admin_tenant_id, image_name):
        try:
            image_id = nova.get_image_id(token, admin_tenant_id, image_name)
            return image_id
        except Exception as e:
            LOG.error(_LE('Failed to get image id for device creation.'
                          ' image name: %(image_name)s. Error: %(error)s'),
                      {'image_name': image_name, 'error': e})

    def create_instance(self, nova, token, admin_tenant_id,
                        image_id, flavor, interfaces_to_attach,
                        instance_name, volume_support,
                        volume_size, files=None, user_data=None,
                        server_grp_id=None):
        try:
            instance_id = nova.create_instance(
                token, admin_tenant_id, image_id, flavor,
                interfaces_to_attach, instance_name, volume_support,
                volume_size, files=files, userdata=user_data,
                server_grp_id=server_grp_id)
            return instance_id
        except Exception as e:
            LOG.error(_LE('Failed to create instance.'
                          'Error: %(error)s'), {'error': e})

    def get_neutron_port_details(self, network_handler, token, port_id):
        try:
            (mgmt_ip_address,
             mgmt_mac, mgmt_cidr, gateway_ip,
             mgmt_port, mgmt_subnet) = (
                network_handler.get_neutron_port_details(token, port_id))

            result = {'neutron_port': mgmt_port['port'],
                      'neutron_subnet': mgmt_subnet['subnet'],
                      'ip_address': mgmt_ip_address,
                      'mac': mgmt_mac,
                      'cidr': mgmt_cidr,
                      'gateway_ip': gateway_ip}
            return result
        except Exception as e:
            import sys
            import traceback
            exc_type, exc_value, exc_traceback = sys.exc_info()
            LOG.error(traceback.format_exception(exc_type, exc_value,
                                                 exc_traceback))
            LOG.error(_LE('Failed to get management port details. '
                          'Error: %(error)s'), {'error': e})

    @_set_network_handler
    def create_network_function_device(self, device_data,
                                       network_handler=None):
        """ Create a NFD

        :param device_data: NFD data
        :type device_data: dict

        :returns: None -- when there is a failure in creating NFD
        :return: dict -- NFD created

        :raises: exceptions.IncompleteData,
                 exceptions.ComputePolicyNotSupported
        """
        self._validate_create_nfd_data(device_data)

        token = device_data['token']
        admin_tenant_id = device_data['admin_tenant_id']
        image_name = self._get_image_name(device_data)

        pre_launch_executor = nfp_executor.TaskExecutor(jobs=3)

        image_id_result = {}
        provider_metadata_result = {}

        pre_launch_executor.add_job('UPDATE_PROVIDER_METADATA',
                                    self._update_provider_metadata_fast,
                                    token, admin_tenant_id,
                                    image_name, device_data,
                                    result_store=provider_metadata_result)
        pre_launch_executor.add_job('GET_INTERFACES_FOR_DEVICE_CREATE',
                                    self._get_interfaces_for_device_create,
                                    token, admin_tenant_id,
                                    network_handler, device_data)
        pre_launch_executor.add_job('GET_IMAGE_ID',
                                    self.get_image_id,
                                    self.compute_handler_nova, token,
                                    admin_tenant_id,
                                    image_name, result_store=image_id_result)

        pre_launch_executor.fire()

        interfaces, image_id, provider_metadata = (
            self._validate_pre_launch_executor_results(
                network_handler,
                device_data,
                image_name,
                image_id_result,
                provider_metadata_result))
        if not interfaces:
            return None

        management_interface = interfaces[0]
        flavor = self._get_service_instance_flavor(device_data)

        interfaces_to_attach = []
        try:
            for interface in interfaces:
                interfaces_to_attach.append({'port': interface['port_id']})
            if provider_metadata.get('supports_hotplug') is False:
                self._update_interfaces_for_non_hotplug_support(
                    network_handler,
                    interfaces,
                    interfaces_to_attach,
                    device_data)
        except Exception as e:
            LOG.error(_LE('Failed to fetch list of interfaces to attach'
                          ' for device creation %(error)s'), {'error': e})
            self._delete_interfaces(device_data, interfaces,
                                    network_handler=network_handler)
            return None

        instance_name = device_data['name']

        create_instance_executor = nfp_executor.TaskExecutor(jobs=3)
        instance_id_result = {}
        port_details_result = {}
        volume_support = device_data['volume_support']
        volume_size = device_data['volume_size']
        create_instance_executor.add_job(
            'CREATE_INSTANCE', self.create_instance,
            self.compute_handler_nova, token,
            admin_tenant_id, image_id, flavor,
            interfaces_to_attach, instance_name,
            volume_support, volume_size,
            files=device_data.get('files'),
            user_data=device_data.get('user_data'),
            result_store=instance_id_result)

        create_instance_executor.add_job(
            'GET_NEUTRON_PORT_DETAILS',
            self.get_neutron_port_details,
            network_handler, token,
            management_interface['port_id'],
            result_store=port_details_result)

        create_instance_executor.fire()

        instance_id, mgmt_neutron_port_info = (
            self._validate_create_instance_executor_results(
                network_handler,
                device_data,
                interfaces,
                instance_id_result,
                port_details_result))
        if not instance_id:
            return None

        mgmt_ip_address = mgmt_neutron_port_info['ip_address']
        return {'id': instance_id,
                'name': instance_name,
                'provider_metadata': provider_metadata,
                'mgmt_ip_address': mgmt_ip_address,
                'mgmt_port_id': interfaces[0],
                'mgmt_neutron_port_info': mgmt_neutron_port_info,
                'max_interfaces': self.maximum_interfaces,
                'interfaces_in_use': len(interfaces_to_attach),
                'description': ''}  # TODO(RPM): what should be the description

    def _validate_create_nfd_data(self, device_data):
        if (
            any(key not in device_data
                for key in ['service_details',
                            'name',
                            'management_network_info',
                            'ports']) or

            type(device_data['service_details']) is not dict or

            any(key not in device_data['service_details']
                for key in ['service_vendor',
                            'device_type',
                            'network_mode']) or

            any(key not in device_data['management_network_info']
                for key in ['id']) or

            type(device_data['ports']) is not list or

            any(key not in port
                for port in device_data['ports']
                for key in ['id',
                            'port_classification',
                            'port_model'])
        ):
            raise exceptions.IncompleteData()

        if (
            device_data['service_details']['device_type'] !=
            nfp_constants.NOVA_MODE
        ):
            raise exceptions.ComputePolicyNotSupported(
                compute_policy=device_data['service_details']['device_type'])

    def _validate_pre_launch_executor_results(self, network_handler,
                                              device_data,
                                              image_name,
                                              image_id_result,
                                              provider_metadata_result,
                                              server_grp_id_result=None):
        interfaces = device_data.pop('interfaces', None)
        if not interfaces:
            LOG.exception(_LE('Failed to get interfaces for device creation.'))
            return None, _, _

        image_id = image_id_result.get('result', None)
        if not image_id:
            LOG.error(_LE('Failed to get image id for device creation.'))
            self._delete_interfaces(device_data, interfaces,
                                    network_handler=network_handler)
            return None, _, _

        if server_grp_id_result and not server_grp_id_result.get('result'):
            LOG.error(_LE('Validation failed for Nova anti-affinity '
                          'server group.'))
            return None, _, _

        provider_metadata = provider_metadata_result.get('result', None)
        if not provider_metadata:
            LOG.warning(_LW('Failed to get provider metadata for'
                            ' device creation.'))
            provider_metadata = {}

        return interfaces, image_id, provider_metadata

    def _get_service_instance_flavor(self, device_data):
        if device_data['service_details'].get('flavor'):
            flavor = device_data['service_details']['flavor']
        else:
            LOG.debug("No Device flavor provided in service profile's "
                      "service flavor field, using default "
                      "flavor: m1.medium")
            flavor = 'm1.medium'
        return flavor

    def _update_interfaces_for_non_hotplug_support(self, network_handler,
                                                   interfaces,
                                                   interfaces_to_attach,
                                                   device_data):
        token = device_data['token']
        enable_port_security = device_data.get('enable_port_security')
        if not device_data['interfaces_to_attach']:
            for port in device_data['ports']:
                if (port['port_classification'] ==
                        nfp_constants.PROVIDER):
                    if (device_data['service_details'][
                        'service_type'].lower()
                        in [nfp_constants.FIREWALL.lower(),
                            nfp_constants.VPN.lower()]):
                        network_handler.set_promiscuos_mode(
                            token, port['id'], enable_port_security)
                    port_id = network_handler.get_port_id(
                        token, port['id'])
                    interfaces_to_attach.append({'port': port_id})
            for port in device_data['ports']:
                if (port['port_classification'] ==
                        nfp_constants.CONSUMER):
                    if (device_data['service_details'][
                        'service_type'].lower()
                        in [nfp_constants.FIREWALL.lower(),
                            nfp_constants.VPN.lower()]):
                        network_handler.set_promiscuos_mode(
                            token, port['id'], enable_port_security)
                    port_id = network_handler.get_port_id(
                        token, port['id'])
                    interfaces_to_attach.append({'port': port_id})
        else:
            for interface in device_data['interfaces_to_attach']:
                interfaces_to_attach.append(
                    {'port': interface['port']})
                interfaces.append({'id': interface['id']})

    def _validate_create_instance_executor_results(self,
                                                   network_handler,
                                                   device_data,
                                                   interfaces,
                                                   instance_id_result,
                                                   port_details_result):
        token = device_data['token']
        admin_tenant_id = device_data['admin_tenant_id']
        instance_id = instance_id_result.get('result', None)
        if not instance_id:
            LOG.error(_LE('Failed to create instance with device data:'
                          '%(data)s.'),
                      {'data': device_data})
            self._delete_interfaces(device_data, interfaces,
                                    network_handler=network_handler)
            return None, _

        mgmt_neutron_port_info = port_details_result.get('result', None)

        if not mgmt_neutron_port_info:
            LOG.error(_LE('Failed to get management port details. '))
            with nfp_ctx_mgr.NovaContextManager as ncm:
                ncm.retry(self.compute_handler_nova.delete_instance,
                          token,
                          admin_tenant_id,
                          instance_id)
            self._delete_interfaces(device_data, interfaces,
                                    network_handler=network_handler)
            return None, _
        return instance_id, mgmt_neutron_port_info

    @_set_network_handler
    def delete_network_function_device(self, device_data,
                                       network_handler=None):
        """ Delete the NFD

        :param device_data: NFD
        :type device_data: dict

        :returns: None -- Both on success and Failure

        :raises: exceptions.IncompleteData,
                 exceptions.ComputePolicyNotSupported
        """
        if (
            any(key not in device_data
                for key in ['service_details',
                            'mgmt_port_id']) or

            type(device_data['service_details']) is not dict or

            any(key not in device_data['service_details']
                for key in ['service_vendor',
                            'device_type',
                            'network_mode']) or

            type(device_data['mgmt_port_id']) is not dict or

            any(key not in device_data['mgmt_port_id']
                for key in ['id',
                            'port_classification',
                            'port_model'])
        ):
            raise exceptions.IncompleteData()

        if (
            device_data['service_details']['device_type'] !=
            nfp_constants.NOVA_MODE
        ):
            raise exceptions.ComputePolicyNotSupported(
                compute_policy=device_data['service_details']['device_type'])

        token = self._get_token(device_data.get('token'))

        if not token:
            return None

        if device_data.get('id'):
            # delete the device instance
            #
            # this method will be invoked again
            # once the device instance deletion is completed
            with nfp_ctx_mgr.NovaContextManager.new(
                    suppress=(Exception,)) as ncm:

                ncm.retry(self.compute_handler_nova.delete_instance,
                          token,
                          device_data['tenant_id'],
                          device_data['id'])
        else:
            # device instance deletion is done, delete remaining resources
            try:
                interfaces = [device_data['mgmt_port_id']]
                self._delete_interfaces(device_data,
                                        interfaces,
                                        network_handler=network_handler)
            except Exception as e:
                LOG.error(_LE('Failed to delete the management data port(s). '
                              'Error: %(error)s'), {'error': e})

    def get_network_function_device_status(self, device_data,
                                           ignore_failure=False):
        """ Get the status of NFD

        :param device_data: NFD
        :type device_data: dict

        :returns: None -- On failure
        :return: str -- status string

        :raises: exceptions.IncompleteData,
                 exceptions.ComputePolicyNotSupported
        """
        if (
            any(key not in device_data
                for key in ['id',
                            'service_details']) or

            type(device_data['service_details']) is not dict or

            any(key not in device_data['service_details']
                for key in ['service_vendor',
                            'device_type',
                            'network_mode'])
        ):
            raise exceptions.IncompleteData()

        if (
            device_data['service_details']['device_type'] !=
            nfp_constants.NOVA_MODE
        ):
            raise exceptions.ComputePolicyNotSupported(
                compute_policy=device_data['service_details']['device_type'])

        token = self._get_token(device_data.get('token'))

        if not token:
            return None

        with nfp_ctx_mgr.NovaContextManager.new(suppress=(Exception,)) as ncm:
            device = ncm.retry(self.compute_handler_nova.get_instance,
                               device_data['token'],
                               device_data['tenant_id'],
                               device_data['id'])

            return device['status']

    @_set_network_handler
    def plug_network_function_device_interfaces(self, device_data,
                                                network_handler=None):
        """ Attach the network interfaces for NFD

        :param device_data: NFD
        :type device_data: dict

        :returns: bool -- False on failure and True on Success

        :raises: exceptions.IncompleteData,
                 exceptions.ComputePolicyNotSupported
        """

        if (
            any(key not in device_data
                for key in ['id',
                            'service_details',
                            'ports']) or

            type(device_data['service_details']) is not dict or

            any(key not in device_data['service_details']
                for key in ['service_vendor',
                            'device_type',
                            'network_mode']) or

            type(device_data['ports']) is not list or

            any(key not in port
                for port in device_data['ports']
                for key in ['id',
                            'port_classification',
                            'port_model'])
        ):
            raise exceptions.IncompleteData()

        if (
            device_data['service_details']['device_type'] !=
            nfp_constants.NOVA_MODE
        ):
            raise exceptions.ComputePolicyNotSupported(
                compute_policy=device_data['service_details']['device_type'])

        token = device_data['token']
        tenant_id = device_data['tenant_id']
        provider_metadata = device_data['provider_metadata']
        enable_port_security = device_data.get('enable_port_security')

        if provider_metadata.get('supports_hotplug') is False:
            return True
        try:
            executor = nfp_executor.TaskExecutor(jobs=10)

            for port in device_data['ports']:
                if port['port_classification'] == nfp_constants.PROVIDER:
                    service_type = device_data[
                        'service_details']['service_type'].lower()
                    if service_type.lower() in [
                            nfp_constants.FIREWALL.lower(),
                            nfp_constants.VPN.lower()]:
                        executor.add_job(
                            'SET_PROMISCUOS_MODE',
                            network_handler.set_promiscuos_mode_fast,
                            token, port['id'], enable_port_security)
                    executor.add_job(
                        'ATTACH_INTERFACE',
                        self.compute_handler_nova.attach_interface,
                        token, tenant_id, device_data['id'],
                        port['id'])
                    break
            executor.fire()

            for port in device_data['ports']:
                if port['port_classification'] == nfp_constants.CONSUMER:
                    service_type = device_data[
                        'service_details']['service_type'].lower()
                    if service_type.lower() in [
                            nfp_constants.FIREWALL.lower(),
                            nfp_constants.VPN.lower()]:
                        executor.add_job(
                            'SET_PROMISCUOS_MODE',
                            network_handler.set_promiscuos_mode_fast,
                            token, port['id'], enable_port_security)
                    executor.add_job(
                        'ATTACH_INTERFACE',
                        self.compute_handler_nova.attach_interface,
                        token, tenant_id, device_data['id'],
                        port['id'])
                    break
            executor.fire()

        except Exception as e:
            LOG.error(_LE('Failed to plug interface(s) to the device.'
                          'Error: %(error)s'), {'error': e})
            return None
        else:
            return True

    @_set_network_handler
    def unplug_network_function_device_interfaces(self, device_data,
                                                  network_handler=None):
        """ Detach the network interfaces for NFD

        :param device_data: NFD
        :type device_data: dict

        :returns: bool -- False on failure and True on Success

        :raises: exceptions.IncompleteData,
                 exceptions.ComputePolicyNotSupported
        """

        if (
            any(key not in device_data
                for key in ['id',
                            'service_details',
                            'ports']) or

            type(device_data['service_details']) is not dict or

            any(key not in device_data['service_details']
                for key in ['service_vendor',
                            'device_type',
                            'network_mode']) or

            any(key not in port
                for port in device_data['ports']
                for key in ['id',
                            'port_classification',
                            'port_model'])
        ):
            raise exceptions.IncompleteData()

        if (
            device_data['service_details']['device_type'] !=
            nfp_constants.NOVA_MODE
        ):
            raise exceptions.ComputePolicyNotSupported(
                compute_policy=device_data['service_details']['device_type'])

        token = self._get_token(device_data.get('token'))

        if not token:
            return None

        image_name = self._get_image_name(device_data)
        provider_metadata = {}
        if image_name:
            provider_metadata = (
                self._update_provider_metadata_fast(token,
                                                    device_data['tenant_id'],
                                                    image_name,
                                                    device_data))

        if not provider_metadata:
            LOG.debug('Failed to get provider metadata for'
                      ' device deletion.')

        if provider_metadata.get('supports_hotplug') is False:
            return True

        with nfp_ctx_mgr.NovaContextManager.new(suppress=(Exception,)) as ncm:
            for port in device_data['ports']:
                port_id = network_handler.get_port_id(token, port['id'])
                ncm.retry(self.compute_handler_nova.detach_interface,
                          token,
                          device_data['tenant_id'],
                          device_data['id'],
                          port_id)
            return True

    @_set_network_handler
    def get_delete_device_data(self, device_data, network_handler=None):
        """ Get the configuration information for NFD

        :param device_data: NFD
        :type device_data: dict

        :returns: None -- On Failure
        :returns: dict

        :raises: exceptions.IncompleteData

        """

        if (
            any(key not in device_data
                for key in ['service_details',
                            'mgmt_ip_address',
                            'ports']) or

            type(device_data['service_details']) is not dict or

            any(key not in device_data['service_details']
                for key in ['service_vendor',
                            'device_type',
                            'network_mode']) or

            type(device_data['ports']) is not list or

            any(key not in port
                for port in device_data['ports']
                for key in ['id',
                            'port_classification',
                            'port_model'])
        ):
            LOG.error(_LE('Incomplete device data received for delete '
                          'network function device.'))
            return None

        token = self._get_token(device_data.get('token'))
        if not token:
            return None

        provider_ip = None
        provider_mac = None
        provider_cidr = None
        consumer_ip = None
        consumer_mac = None
        consumer_cidr = None
        consumer_gateway_ip = None

        for port in device_data['ports']:
            if port['port_classification'] == nfp_constants.PROVIDER:
                try:
                    (provider_ip, provider_mac, provider_cidr, dummy, _, _) = (
                        network_handler.get_port_details(token, port['id'])
                    )
                except Exception:
                    LOG.error(_LE('Failed to get provider port details'
                                  ' for get device config info operation'))
                    return None
            elif port['port_classification'] == nfp_constants.CONSUMER:
                try:
                    (consumer_ip, consumer_mac, consumer_cidr,
                     consumer_gateway_ip, _, _) = (
                        network_handler.get_port_details(token, port['id'])
                    )
                except Exception:
                    LOG.error(_LE('Failed to get consumer port details'
                                  ' for get device config info operation'))
                    return None

        device_data.update({
            'provider_ip': provider_ip, 'provider_mac': provider_mac,
            'provider_cidr': provider_cidr, 'consumer_ip': consumer_ip,
            'consumer_mac': consumer_mac, 'consumer_cidr': consumer_cidr,
            'consumer_gateway_ip': consumer_gateway_ip})

        return device_data

    @_set_network_handler
    def get_network_function_device_config(self, device_data,
                                           resource_type, is_delete=False,
                                           network_handler=None):
        """ Get the configuration information for NFD

        :returns: dict

        """

        if is_delete:
            device_data = self.get_delete_device_data(
                device_data, network_handler=network_handler)
            if not device_data:
                return None

        return df.get_network_function_info(
            device_data, resource_type)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import six

from gbpservice.nfp.core import log as nfp_logging

LOG = nfp_logging.getLogger(__name__)

deque = collections.deque


class SequencerEmpty(Exception):
    pass


class SequencerBusy(Exception):
    pass

"""Sequences the events. """


class EventSequencer(object):

    class Sequencer(object):

        def __init__(self):
            # Events not scheduled are queued
            self._waitq = deque()
            # Currently scheduled event
            self._scheduled = None

        def _is_busy(self):
            if self._scheduled:
                raise SequencerBusy

        def _is_empty(self):
            if not len(self._waitq):
                raise SequencerEmpty

        def sequence(self, event):
            self._waitq.append(event)

        def run(self):
            """Run to get event to be scheduled.

                If sequencer is busy - i.e, an event is already
                scheduled and in progress raises busy except.
                If sequencer is empty - i.e, no event in sequencer
                raises empty except.
            """
            self._is_busy()
            self._is_empty()
            # Pop the first element in the queue - FIFO
            self._scheduled = self._waitq.popleft()
            return self._scheduled

        def is_scheduled(self, event):
            if self._scheduled:
                return self._scheduled.desc.uuid == event.desc.uuid and (
                    self._scheduled.id == event.id)
            return True

        def release(self):
            self._scheduled = None

        def pop(self):
            self.release()
            events = list(self._waitq)
            self._waitq.clear()
            return events

    def __init__(self):
        # Sequence of related events
        # {key: sequencer()}
        self._sequencer = {}

    def sequence(self, key, event):
        try:
            self._sequencer[key].sequence(event)
        except KeyError:
            self._sequencer[key] = self.Sequencer()
            self._sequencer[key].sequence(event)
        message = "Sequenced event - %s" % (event.identify())
        LOG.debug(message)

    def run(self):
        events = []
        # Loop over copy and delete from original
        sequencers = dict(self._sequencer)
        for key, sequencer in six.iteritems(sequencers):
            try:
                event = sequencer.run()
                if event:
                    message = "Desequenced event - %s" % (
                        event.identify())
                    LOG.debug(message)
                    event.sequence = False
                    events.append(event)
            except SequencerBusy as exc:
                pass
            except SequencerEmpty as exc:
                exc = exc
                message = "Sequencer empty"
                LOG.debug(message)
                del self._sequencer[key]
        return events

    def pop(self):
        events = []
        sequencers = dict(self._sequencer)
        for key, sequencer in six.iteritems(sequencers):
            events += sequencer.pop()
        return events

    def release(self, key, event):
        try:
            message = "(event - %s) checking to release" % (event.identify())
            LOG.debug(message)
            if self._sequencer[key].is_scheduled(event):
                message = "(event - %s) Releasing sequencer" % (
                    event.identify())
                LOG.debug(message)
                self._sequencer[key].release()
        except KeyError:
            return



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import commands
import subprocess
import sys


def main():
    """
    Main: Wrapper for shared_func tests
    """
    # Usage: python suite_admin_run.py
    print ("Functional Test Script to execute Shared Resource Testcases")
    cmd_list = ["sudo sh -c 'cat /dev/null > test_results_admin.txt'",
                "sudo chmod 777 test_results_admin.txt "]
    for cmd in cmd_list:
        commands.getoutput(cmd)
    test_list = ['tc_gbp_pr_pc_pa_shared_func.py',
                 'tc_gbp_prs_pr_shared_func.py']
    for test in test_list:
        cmd = 'python %s' % (test)
        print (cmd)
        subprocess.call(cmd, shell=True)
    results_file = open("test_results_admin.txt")
    contents = results_file.read()
    results_file.close()
    print (contents)
    print ("\n\nTotal Number of Shared Resource TestCases Executed= %s" % (
        contents.count("_SHARED_")))
    print ("\n\nNumber of TestCases Passed= %s" % (contents.count("PASSED")))
    print ("\n\nNumber of TestCases Failed= %s" % (contents.count("FAILED")))
    if contents.count("FAILED") > 0:
        sys.exit(1)
    else:
        return 0

if __name__ == "__main__":
    main()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import log as logging

from neutron.extensions import portbindings
from neutron.services.trunk import constants as trunk_consts
from neutron.services.trunk.drivers import base

from opflexagent import constants as ofcst


LOG = logging.getLogger(__name__)

NAME = 'apic_aim'

SUPPORTED_INTERFACES = (
    portbindings.VIF_TYPE_OVS,
    portbindings.VIF_TYPE_VHOST_USER,
)

SUPPORTED_SEGMENTATION_TYPES = (
    trunk_consts.VLAN,
)

DRIVER = None


class OpflexDriver(base.DriverBase):

    @property
    def is_loaded(self):
        try:
            return NAME in cfg.CONF.ml2.mechanism_drivers
        except cfg.NoSuchOptError:
            return False

    @classmethod
    def create(cls):
        return OpflexDriver(NAME,
                            SUPPORTED_INTERFACES,
                            SUPPORTED_SEGMENTATION_TYPES,
                            ofcst.AGENT_TYPE_OPFLEX_OVS)


def register():
    """Register the driver."""
    global DRIVER
    DRIVER = OpflexDriver.create()
    LOG.debug('Opflex trunk driver registered')
